"""Cross-currency basis-swap pricing helpers.

PVs are reported in the chosen valuation currency. The par spread is a raw
decimal and FX conversions follow the swap's pay and receive currencies.
"""

from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal

from fuggers_py.core.types import Currency, Date
from fuggers_py.data.ids import CurrencyPair
from fuggers_py.data.pricing_specs import AnalyticsCurves
from fuggers_py.rates.instruments import CrossCurrencyBasisSwap, FloatingLegSpec, PayReceive

from ._curve_resolver import forward_rate_from_curve, resolve_discount_curve, resolve_projection_curve


def _to_decimal(value: object) -> Decimal:
    if isinstance(value, Decimal):
        return value
    return Decimal(str(value))


def _pair(base: Currency, quote: Currency) -> CurrencyPair:
    return CurrencyPair(base=base, quote=quote)


def _curve_value_at_date(candidate: object, date: Date) -> Decimal | None:
    if hasattr(candidate, "value_at_date"):
        return _to_decimal(candidate.value_at_date(date))
    if hasattr(candidate, "date_to_tenor") and hasattr(candidate, "value_at"):
        tenor = candidate.date_to_tenor(date)
        return _to_decimal(candidate.value_at(tenor))
    return None


def _call_fx_method(curve: object, pair: CurrencyPair, date: Date) -> Decimal | None:
    for method_name in ("forward_rate", "fx_forward_rate", "forward_fx_rate", "rate", "get_forward_rate"):
        method = getattr(curve, method_name, None)
        if method is None:
            continue
        attempts = (
            lambda: method(pair, date),
            lambda: method(pair.as_str(), date),
            lambda: method(currency_pair=pair, date=date),
            lambda: method(currency_pair=pair.as_str(), date=date),
        )
        for attempt in attempts:
            try:
                value = attempt()
            except TypeError:
                continue
            except Exception:
                continue
            if value is not None:
                return _to_decimal(value)
    return None


def _forward_rate_from_explicit_curve(curves: AnalyticsCurves, pair: CurrencyPair, date: Date) -> Decimal | None:
    curve = curves.fx_forward_curve
    if curve is None:
        return None
    direct = _call_fx_method(curve, pair, date)
    if direct is not None:
        return direct

    inverse_pair = _pair(pair.quote, pair.base)
    inverse = _call_fx_method(curve, inverse_pair, date)
    if inverse is not None:
        if inverse == Decimal(0):
            raise ValueError("FX forward curve returned a zero inverse forward rate.")
        return Decimal(1) / inverse

    if isinstance(curve, dict) or hasattr(curve, "get"):
        getter = curve.get if hasattr(curve, "get") else curve.__getitem__
        for candidate_pair, invert in ((pair, False), (inverse_pair, True)):
            keys = (candidate_pair, candidate_pair.as_str(), candidate_pair.as_str().lower(), candidate_pair.as_str().upper())
            for key in keys:
                try:
                    candidate = getter(key)
                except Exception:
                    continue
                if candidate is None:
                    continue
                if isinstance(candidate, (Decimal, int, float, str)):
                    value = _to_decimal(candidate)
                else:
                    value = _call_fx_method(candidate, candidate_pair, date)
                    if value is None:
                        value = _curve_value_at_date(candidate, date)
                if value is None:
                    continue
                if invert:
                    if value == Decimal(0):
                        raise ValueError("FX forward sub-curve returned a zero inverse forward rate.")
                    return Decimal(1) / value
                return value
    explicit_value = _curve_value_at_date(curve, date)
    if explicit_value is not None:
        return explicit_value
    return None


@dataclass(frozen=True, slots=True)
class CrossCurrencyBasisSwapPricingResult:
    """Cross-currency basis-swap pricing output."""

    valuation_currency: Currency
    par_spread: Decimal
    present_value: Decimal
    pay_leg_pv: Decimal
    receive_leg_pv: Decimal
    principal_exchange_pv: Decimal
    spread_annuity: Decimal


@dataclass(frozen=True, slots=True)
class CrossCurrencyBasisSwapPricer:
    """Price cross-currency basis swaps."""

    def _valuation_currency(
        self,
        swap: CrossCurrencyBasisSwap,
        valuation_currency: Currency | None,
    ) -> Currency:
        resolved = valuation_currency or swap.pay_leg.currency
        if resolved not in {swap.pay_leg.currency, swap.receive_leg.currency}:
            raise ValueError("valuation_currency must be one of the swap leg currencies.")
        return resolved

    def _pair_forward_rate(self, swap: CrossCurrencyBasisSwap, curves: AnalyticsCurves, date: Date) -> Decimal:
        pair = swap.currency_pair()
        explicit = _forward_rate_from_explicit_curve(curves, pair, date)
        if explicit is not None:
            return explicit
        pay_discount_curve = resolve_discount_curve(curves, swap.pay_leg.currency)
        receive_discount_curve = resolve_discount_curve(curves, swap.receive_leg.currency)
        return swap.spot_fx_rate * pay_discount_curve.discount_factor(date) / receive_discount_curve.discount_factor(date)

    def _fx_conversion_rate(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        source_currency: Currency,
        target_currency: Currency,
        date: Date,
    ) -> Decimal:
        if source_currency is target_currency:
            return Decimal(1)
        forward = self._pair_forward_rate(swap, curves, date)
        if source_currency is swap.pay_leg.currency and target_currency is swap.receive_leg.currency:
            return forward
        if source_currency is swap.receive_leg.currency and target_currency is swap.pay_leg.currency:
            if forward == Decimal(0):
                raise ValueError("CrossCurrencyBasisSwap requires a non-zero FX forward rate.")
            return Decimal(1) / forward
        raise ValueError("FX conversion only supports the swap's two leg currencies.")

    def _leg_pv(
        self,
        swap: CrossCurrencyBasisSwap,
        leg: FloatingLegSpec,
        periods,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency,
        include_spread: bool,
    ) -> Decimal:
        projection_curve = resolve_projection_curve(
            curves,
            currency=leg.currency,
            index_name=leg.index_name,
            index_tenor=leg.index_tenor,
        )
        discount_curve = resolve_discount_curve(curves, valuation_currency)
        spread = leg.spread if include_spread else Decimal(0)
        present_value = Decimal(0)
        sign = leg.pay_receive.sign()
        for period in periods:
            forward = forward_rate_from_curve(
                projection_curve,
                period.start_date,
                period.end_date,
                day_count_convention=leg.day_count_convention,
            )
            coupon = leg.notional * (forward + spread) * period.year_fraction
            fx_conversion = self._fx_conversion_rate(
                swap,
                curves,
                source_currency=leg.currency,
                target_currency=valuation_currency,
                date=period.payment_date,
            )
            present_value += sign * coupon * fx_conversion * discount_curve.discount_factor(period.payment_date)
        return present_value

    def _principal_exchange_pv(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency,
    ) -> Decimal:
        discount_curve = resolve_discount_curve(curves, valuation_currency)
        present_value = Decimal(0)
        if swap.initial_exchange:
            for leg in (swap.pay_leg, swap.receive_leg):
                fx_conversion = self._fx_conversion_rate(
                    swap,
                    curves,
                    source_currency=leg.currency,
                    target_currency=valuation_currency,
                    date=swap.effective_date,
                )
                present_value += leg.pay_receive.sign() * leg.notional * fx_conversion * discount_curve.discount_factor(
                    swap.effective_date
                )
        if swap.final_exchange:
            for leg in (swap.pay_leg, swap.receive_leg):
                fx_conversion = self._fx_conversion_rate(
                    swap,
                    curves,
                    source_currency=leg.currency,
                    target_currency=valuation_currency,
                    date=swap.maturity_date,
                )
                present_value -= leg.pay_receive.sign() * leg.notional * fx_conversion * discount_curve.discount_factor(
                    swap.maturity_date
                )
        return present_value

    def _annuity(
        self,
        swap: CrossCurrencyBasisSwap,
        leg: FloatingLegSpec,
        periods,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency,
    ) -> Decimal:
        discount_curve = resolve_discount_curve(curves, valuation_currency)
        annuity = Decimal(0)
        for period in periods:
            fx_conversion = self._fx_conversion_rate(
                swap,
                curves,
                source_currency=leg.currency,
                target_currency=valuation_currency,
                date=period.payment_date,
            )
            annuity += leg.notional * period.year_fraction * fx_conversion * discount_curve.discount_factor(period.payment_date)
        if annuity == Decimal(0):
            raise ValueError("CrossCurrencyBasisSwap par spread requires a non-zero quoted-leg annuity.")
        return annuity

    def pay_leg_pv(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency | None = None,
    ) -> Decimal:
        """Return the discounted PV of the pay leg in the valuation currency."""

        resolved_currency = self._valuation_currency(swap, valuation_currency)
        return self._leg_pv(
            swap,
            swap.pay_leg,
            swap.pay_periods(),
            curves,
            valuation_currency=resolved_currency,
            include_spread=True,
        )

    def receive_leg_pv(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency | None = None,
    ) -> Decimal:
        """Return the discounted PV of the receive leg in the valuation currency."""

        resolved_currency = self._valuation_currency(swap, valuation_currency)
        return self._leg_pv(
            swap,
            swap.receive_leg,
            swap.receive_periods(),
            curves,
            valuation_currency=resolved_currency,
            include_spread=True,
        )

    def principal_exchange_pv(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency | None = None,
    ) -> Decimal:
        """Return the PV of the principal exchanges."""

        return self._principal_exchange_pv(
            swap,
            curves,
            valuation_currency=self._valuation_currency(swap, valuation_currency),
        )

    def pv(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency | None = None,
    ) -> Decimal:
        """Return the total present value in the valuation currency."""

        resolved_currency = self._valuation_currency(swap, valuation_currency)
        return (
            self.pay_leg_pv(swap, curves, valuation_currency=resolved_currency)
            + self.receive_leg_pv(swap, curves, valuation_currency=resolved_currency)
            + self._principal_exchange_pv(swap, curves, valuation_currency=resolved_currency)
        )

    def par_spread(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency | None = None,
    ) -> Decimal:
        """Return the par spread on the quoted leg as a raw decimal."""

        resolved_currency = self._valuation_currency(swap, valuation_currency)
        principal_exchange_pv = self._principal_exchange_pv(swap, curves, valuation_currency=resolved_currency)
        if swap.quoted_leg is PayReceive.RECEIVE:
            quoted_leg = swap.receive_leg
            quoted_periods = swap.receive_periods()
            other_leg_pv = self._leg_pv(
                swap,
                swap.pay_leg,
                swap.pay_periods(),
                curves,
                valuation_currency=resolved_currency,
                include_spread=True,
            )
        else:
            quoted_leg = swap.pay_leg
            quoted_periods = swap.pay_periods()
            other_leg_pv = self._leg_pv(
                swap,
                swap.receive_leg,
                swap.receive_periods(),
                curves,
                valuation_currency=resolved_currency,
                include_spread=True,
            )
        quoted_leg_pv_without_spread = self._leg_pv(
            swap,
            quoted_leg,
            quoted_periods,
            curves,
            valuation_currency=resolved_currency,
            include_spread=False,
        )
        annuity = self._annuity(
            swap,
            quoted_leg,
            quoted_periods,
            curves,
            valuation_currency=resolved_currency,
        )
        return -(other_leg_pv + quoted_leg_pv_without_spread + principal_exchange_pv) / (
            quoted_leg.pay_receive.sign() * annuity
        )

    def price(
        self,
        swap: CrossCurrencyBasisSwap,
        curves: AnalyticsCurves,
        *,
        valuation_currency: Currency | None = None,
    ) -> CrossCurrencyBasisSwapPricingResult:
        """Return the full cross-currency basis-swap pricing result."""

        resolved_currency = self._valuation_currency(swap, valuation_currency)
        pay_leg_pv = self.pay_leg_pv(swap, curves, valuation_currency=resolved_currency)
        receive_leg_pv = self.receive_leg_pv(swap, curves, valuation_currency=resolved_currency)
        principal_exchange_pv = self._principal_exchange_pv(swap, curves, valuation_currency=resolved_currency)
        if swap.quoted_leg is PayReceive.RECEIVE:
            quoted_leg = swap.receive_leg
            quoted_periods = swap.receive_periods()
        else:
            quoted_leg = swap.pay_leg
            quoted_periods = swap.pay_periods()
        spread_annuity = self._annuity(
            swap,
            quoted_leg,
            quoted_periods,
            curves,
            valuation_currency=resolved_currency,
        )
        par_spread = self.par_spread(swap, curves, valuation_currency=resolved_currency)
        return CrossCurrencyBasisSwapPricingResult(
            valuation_currency=resolved_currency,
            par_spread=par_spread,
            present_value=pay_leg_pv + receive_leg_pv + principal_exchange_pv,
            pay_leg_pv=pay_leg_pv,
            receive_leg_pv=receive_leg_pv,
            principal_exchange_pv=principal_exchange_pv,
            spread_annuity=spread_annuity,
        )


__all__ = ["CrossCurrencyBasisSwapPricer", "CrossCurrencyBasisSwapPricingResult"]



"""Rates options tools.

The subpackage covers futures options, swaptions, cap/floor products, and the
associated Black-76, Bachelier, and Hull-White style pricers. Option inputs
are raw-decimal rates or volatilities, while pricer outputs are currency
present values and greeks scaled to the relevant contract notionals.
"""

from __future__ import annotations

from ._common import OptionType
from .cap_floor import CapFloor, CapFloorType
from .futures_option import FuturesOption
from .pricing import (
    BachelierPricer,
    Black76Pricer,
    CapFloorPricingResult,
    CapFloorletPricingResult,
    FuturesOptionPricingResult,
    HullWhiteOptionPricer,
    HullWhiteRateOptionModel,
    OptionFormulaResult,
    OptionGreeks,
    SwaptionPricingResult,
    bachelier_formula,
    black76_formula,
    hull_white_normal_volatility,
)
from .swaption import Swaption

__all__ = [
    "BachelierPricer",
    "Black76Pricer",
    "CapFloor",
    "CapFloorPricingResult",
    "CapFloorType",
    "CapFloorletPricingResult",
    "FuturesOption",
    "FuturesOptionPricingResult",
    "HullWhiteOptionPricer",
    "HullWhiteRateOptionModel",
    "OptionFormulaResult",
    "OptionGreeks",
    "OptionType",
    "Swaption",
    "SwaptionPricingResult",
    "bachelier_formula",
    "black76_formula",
    "hull_white_normal_volatility",
]



from __future__ import annotations

from decimal import Decimal

from fuggers_py.funding import (
    all_in_financing_cost,
    financed_cash,
    haircut_amount,
    haircut_drag,
    haircut_financing_cost,
)


def test_haircut_financing_helpers_compute_amounts_costs_and_drag() -> None:
    collateral_value = Decimal("1000000")
    haircut = Decimal("0.02")
    year_fraction = Decimal("0.25")

    assert haircut_amount(collateral_value=collateral_value, haircut=haircut) == Decimal("20000.00")
    assert financed_cash(collateral_value=collateral_value, haircut=haircut) == Decimal("980000.00")
    assert haircut_financing_cost(
        collateral_value=collateral_value,
        haircut=haircut,
        funding_rate=Decimal("0.06"),
        year_fraction=year_fraction,
    ) == Decimal("300.00000")
    assert all_in_financing_cost(
        collateral_value=collateral_value,
        haircut=haircut,
        repo_rate=Decimal("0.04"),
        haircut_funding_rate=Decimal("0.06"),
        year_fraction=year_fraction,
    ) == Decimal("10100.00000")
    assert haircut_drag(
        collateral_value=collateral_value,
        haircut=haircut,
        repo_rate=Decimal("0.04"),
        haircut_funding_rate=Decimal("0.06"),
        year_fraction=year_fraction,
    ) == Decimal("100.00000")



"""Async market-data fanout and graph integration helpers.

The listener stores incoming market-data payloads, updates the calculation
graph, and marks dependent nodes dirty so the reactive engine can refresh
pricing outputs deterministically.
"""

from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from datetime import UTC, datetime

from fuggers_py.data import (
    CurveId,
    CurveInputs,
    CurrencyPair,
    EtfQuote,
    FxRate,
    IndexFixing,
    InflationFixing,
    InMemoryCurveSource,
    InMemoryFixingSource,
    InMemoryFxRateSource,
    InMemoryInflationFixingSource,
    InMemoryQuoteSource,
    InMemoryVolatilitySource,
    InstrumentId,
    RawQuote,
    VolSurfaceId,
    VolatilitySurface,
    YearMonth,
)

from .calc_graph import CalculationGraph, NodeId
from .curve_builder import CurveBuilder
from .scheduler import NodeUpdate, ThrottleManager, UpdateSource


def _now() -> datetime:
    return datetime.now(UTC)


@dataclass(frozen=True, slots=True, kw_only=True)
class MarketDataUpdate:
    """Base record for a market-data event."""

    timestamp: datetime = field(default_factory=_now)
    source: str | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class QuoteUpdate(MarketDataUpdate):
    """Update carrying a new quote."""

    quote: RawQuote


@dataclass(frozen=True, slots=True, kw_only=True)
class CurveInputUpdate(MarketDataUpdate):
    """Update carrying new curve inputs."""

    curve_inputs: CurveInputs


@dataclass(frozen=True, slots=True, kw_only=True)
class CurveUpdate(MarketDataUpdate):
    """Update carrying a built curve or curve placeholder."""

    curve_id: CurveId
    curve: object | None = None
    curve_inputs: CurveInputs | None = None


@dataclass(frozen=True, slots=True, kw_only=True)
class IndexFixingUpdate(MarketDataUpdate):
    """Update carrying an index fixing."""

    fixing: IndexFixing


@dataclass(frozen=True, slots=True, kw_only=True)
class InflationFixingUpdate(MarketDataUpdate):
    """Update carrying an inflation fixing."""

    fixing: InflationFixing


@dataclass(frozen=True, slots=True, kw_only=True)
class FxRateUpdate(MarketDataUpdate):
    """Update carrying an FX rate."""

    fx_rate: FxRate


@dataclass(frozen=True, slots=True, kw_only=True)
class VolSurfaceUpdate(MarketDataUpdate):
    """Update carrying a volatility surface."""

    surface: VolatilitySurface


class MarketDataPublisher:
    """Async fanout helper for broadcasting market-data updates."""

    def __init__(self) -> None:
        self._subscribers: set[asyncio.Queue] = set()

    def subscribe(self) -> asyncio.Queue:
        """Create and register a subscriber queue."""
        queue: asyncio.Queue = asyncio.Queue()
        self._subscribers.add(queue)
        return queue

    def unsubscribe(self, queue: asyncio.Queue) -> None:
        """Remove a previously registered subscriber queue."""
        self._subscribers.discard(queue)

    async def publish(self, update: MarketDataUpdate) -> None:
        """Publish an update to all current subscribers."""
        for queue in tuple(self._subscribers):
            await queue.put(update)


@dataclass(slots=True)
class MarketDataListener:
    """Translate market-data updates into graph updates and cache writes."""

    calc_graph: CalculationGraph
    curve_builder: CurveBuilder | None = None
    quote_source: InMemoryQuoteSource = field(default_factory=InMemoryQuoteSource)
    curve_source: InMemoryCurveSource = field(default_factory=InMemoryCurveSource)
    fixing_source: InMemoryFixingSource = field(default_factory=InMemoryFixingSource)
    fx_rate_source: InMemoryFxRateSource = field(default_factory=InMemoryFxRateSource)
    inflation_fixing_source: InMemoryInflationFixingSource = field(default_factory=InMemoryInflationFixingSource)
    volatility_source: InMemoryVolatilitySource = field(default_factory=InMemoryVolatilitySource)
    throttle_manager: ThrottleManager | None = None

    @staticmethod
    def quote_node_id(instrument_id: InstrumentId | str) -> NodeId:
        """Return the graph node identifier for a quote."""
        return NodeId.parse(f"quote:{InstrumentId.parse(instrument_id).as_str()}")

    @staticmethod
    def curve_input_node_id(curve_id: CurveId | str) -> NodeId:
        """Return the graph node identifier for curve inputs."""
        return NodeId.parse(f"curve_input:{CurveId.parse(curve_id).as_str()}")

    @staticmethod
    def curve_node_id(curve_id: CurveId | str) -> NodeId:
        """Return the graph node identifier for a built curve."""
        return NodeId.parse(f"curve:{CurveId.parse(curve_id).as_str()}")

    @staticmethod
    def fixing_node_id(index_name: str, fixing_date) -> NodeId:
        """Return the graph node identifier for a fixing."""
        return NodeId.parse(f"fixing:{index_name.strip().upper()}:{fixing_date}")

    @staticmethod
    def inflation_node_id(index_name: str, observation_month: YearMonth | str) -> NodeId:
        """Return the graph node identifier for an inflation fixing."""
        return NodeId.parse(f"inflation:{index_name.strip().upper()}:{YearMonth.parse(observation_month).as_str()}")

    @staticmethod
    def fx_node_id(currency_pair: CurrencyPair | str) -> NodeId:
        """Return the graph node identifier for an FX rate."""
        return NodeId.parse(f"fx:{CurrencyPair.parse(currency_pair).as_str()}")

    @staticmethod
    def vol_surface_node_id(surface_id: VolSurfaceId | str) -> NodeId:
        """Return the graph node identifier for a volatility surface."""
        return NodeId.parse(f"vol:{VolSurfaceId.parse(surface_id).as_str()}")

    async def handle_update(self, update: MarketDataUpdate) -> tuple[NodeUpdate, ...]:
        """Apply a market-data update to the caches and dependency graph."""
        if isinstance(update, QuoteUpdate):
            return self._handle_quote_update(update)
        if isinstance(update, CurveInputUpdate):
            return self._handle_curve_input_update(update)
        if isinstance(update, CurveUpdate):
            return self._handle_curve_update(update)
        if isinstance(update, IndexFixingUpdate):
            return self._handle_fixing_update(update)
        if isinstance(update, InflationFixingUpdate):
            return self._handle_inflation_update(update)
        if isinstance(update, FxRateUpdate):
            return self._handle_fx_update(update)
        if isinstance(update, VolSurfaceUpdate):
            return self._handle_vol_surface_update(update)
        return ()

    async def handle_scheduler_update(self, update: NodeUpdate) -> tuple[NodeUpdate, ...]:
        """Mark a scheduler-triggered node and its dependents dirty."""
        self.calc_graph.add_node(update.node_id)
        dirty = self.calc_graph.mark_dirty(update.node_id, include_self=True)
        return tuple(
            NodeUpdate(node_id=node_id, source=update.source, timestamp=update.timestamp, payload=update.payload, detail=update.detail)
            for node_id in dirty
        )

    async def drain_once(self, queue: asyncio.Queue) -> tuple[NodeUpdate, ...]:
        """Drain a single item from a queue and route it through the listener."""
        item = await queue.get()
        if isinstance(item, NodeUpdate):
            return await self.handle_scheduler_update(item)
        return await self.handle_update(item)

    def _allow(self, key: str, timestamp: datetime) -> bool:
        if self.throttle_manager is None:
            return True
        return self.throttle_manager.allow(key, now=timestamp)

    def _handle_quote_update(self, update: QuoteUpdate) -> tuple[NodeUpdate, ...]:
        node_id = self.quote_node_id(update.quote.instrument_id)
        if not self._allow(node_id.as_str(), update.timestamp):
            return ()
        self.quote_source.add_quote(update.quote)
        self.calc_graph.add_node(node_id)
        self.calc_graph.update_node_value(node_id, update.quote, source=update.source or "market_data", mark_clean=True)
        dirty = self.calc_graph.mark_dependents_dirty(node_id)
        return self._updates((node_id, *dirty), UpdateSource.MARKET_DATA, update)

    def _handle_curve_input_update(self, update: CurveInputUpdate) -> tuple[NodeUpdate, ...]:
        input_node = self.curve_input_node_id(update.curve_inputs.curve_id)
        if not self._allow(input_node.as_str(), update.timestamp):
            return ()
        self.curve_source.add_curve_inputs(update.curve_inputs)
        self.calc_graph.add_node(input_node)
        self.calc_graph.update_node_value(input_node, update.curve_inputs, source=update.source or "market_data", mark_clean=True)
        dirty = list(self.calc_graph.mark_dependents_dirty(input_node))
        touched = [input_node]
        if self.curve_builder is not None:
            built_curve = self.curve_builder.add_from_inputs(update.curve_inputs)
            curve_node = self.curve_node_id(update.curve_inputs.curve_id)
            self.calc_graph.add_node(curve_node)
            self.calc_graph.update_node_value(curve_node, built_curve, source="curve_builder", mark_clean=True)
            dirty.extend(self.calc_graph.mark_dependents_dirty(curve_node))
            touched.append(curve_node)
        return self._updates(tuple(dict.fromkeys((*touched, *dirty))), UpdateSource.MARKET_DATA, update)

    def _handle_curve_update(self, update: CurveUpdate) -> tuple[NodeUpdate, ...]:
        node_id = self.curve_node_id(update.curve_id)
        if not self._allow(node_id.as_str(), update.timestamp):
            return ()
        self.calc_graph.add_node(node_id)
        self.calc_graph.update_node_value(node_id, update.curve, source=update.source or "market_data", mark_clean=True)
        dirty = self.calc_graph.mark_dependents_dirty(node_id)
        return self._updates((node_id, *dirty), UpdateSource.MARKET_DATA, update)

    def _handle_fixing_update(self, update: IndexFixingUpdate) -> tuple[NodeUpdate, ...]:
        node_id = self.fixing_node_id(update.fixing.index_name, update.fixing.fixing_date)
        if not self._allow(node_id.as_str(), update.timestamp):
            return ()
        self.fixing_source.add_fixing(update.fixing)
        self.calc_graph.add_node(node_id)
        self.calc_graph.update_node_value(node_id, update.fixing, source=update.source or "market_data", mark_clean=True)
        dirty = self.calc_graph.mark_dependents_dirty(node_id)
        return self._updates((node_id, *dirty), UpdateSource.MARKET_DATA, update)

    def _handle_inflation_update(self, update: InflationFixingUpdate) -> tuple[NodeUpdate, ...]:
        node_id = self.inflation_node_id(update.fixing.index_name, update.fixing.observation_month)
        if not self._allow(node_id.as_str(), update.timestamp):
            return ()
        self.inflation_fixing_source.add_inflation_fixing(update.fixing)
        self.calc_graph.add_node(node_id)
        self.calc_graph.update_node_value(node_id, update.fixing, source=update.source or "market_data", mark_clean=True)
        dirty = self.calc_graph.mark_dependents_dirty(node_id)
        return self._updates((node_id, *dirty), UpdateSource.MARKET_DATA, update)

    def _handle_fx_update(self, update: FxRateUpdate) -> tuple[NodeUpdate, ...]:
        node_id = self.fx_node_id(update.fx_rate.currency_pair)
        if not self._allow(node_id.as_str(), update.timestamp):
            return ()
        self.fx_rate_source.add_fx_rate(update.fx_rate)
        self.calc_graph.add_node(node_id)
        self.calc_graph.update_node_value(node_id, update.fx_rate, source=update.source or "market_data", mark_clean=True)
        dirty = self.calc_graph.mark_dependents_dirty(node_id)
        return self._updates((node_id, *dirty), UpdateSource.MARKET_DATA, update)

    def _handle_vol_surface_update(self, update: VolSurfaceUpdate) -> tuple[NodeUpdate, ...]:
        node_id = self.vol_surface_node_id(update.surface.surface_id)
        if not self._allow(node_id.as_str(), update.timestamp):
            return ()
        self.volatility_source.add_surface(update.surface)
        self.calc_graph.add_node(node_id)
        self.calc_graph.update_node_value(node_id, update.surface, source=update.source or "market_data", mark_clean=True)
        dirty = self.calc_graph.mark_dependents_dirty(node_id)
        return self._updates((node_id, *dirty), UpdateSource.MARKET_DATA, update)

    def _updates(self, node_ids: tuple[NodeId, ...], source: UpdateSource, payload: object) -> tuple[NodeUpdate, ...]:
        return tuple(
            NodeUpdate(node_id=node_id, source=source, timestamp=getattr(payload, "timestamp", _now()), payload=payload)
            for node_id in node_ids
        )


__all__ = [
    "CurveInputUpdate",
    "CurveUpdate",
    "FxRateUpdate",
    "IndexFixingUpdate",
    "InflationFixingUpdate",
    "MarketDataListener",
    "MarketDataPublisher",
    "MarketDataUpdate",
    "QuoteUpdate",
    "VolSurfaceUpdate",
]



"""Derived curves built from a base curve and simple transforms.

Transforms operate on continuously compounded zero rates in raw decimal form.
This makes parallel shifts, multiplicative overlays, and additive spread
overlays explicit and easy to combine.
"""

from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Callable

from fuggers_py.core.traits import YieldCurve
from fuggers_py.core.types import Compounding, Date, Yield

from ..conversion import ValueConverter
from ..wrappers import RateCurve


def _to_decimal(value: object) -> Decimal:
    if isinstance(value, Decimal):
        return value
    return Decimal(str(value))


class CurveTransformKind(str, Enum):
    """Kinds of transform supported by :class:`CurveTransform`."""

    PARALLEL_SHIFT = "PARALLEL_SHIFT"
    MULTIPLICATIVE = "MULTIPLICATIVE"
    SPREAD_OVERLAY = "SPREAD_OVERLAY"
    FUNCTION = "FUNCTION"


@dataclass(frozen=True, slots=True)
class CurveTransform:
    """Single transform applied to a base zero rate.

    Parameters
    ----------
    kind:
        Transform family.
    amount:
        Raw decimal amount used by parallel-shift and multiplicative
        transforms.
    overlay:
        Optional curve whose continuously compounded zero rate is added for
        spread overlays.
    fn:
        Optional callable used for arbitrary transforms.
    """

    kind: CurveTransformKind
    amount: Decimal = Decimal(0)
    overlay: YieldCurve | None = None
    fn: Callable[[float, float], float] | None = None

    @classmethod
    def parallel_shift(cls, amount: object) -> "CurveTransform":
        """Create an additive raw-decimal zero-rate shift."""
        return cls(kind=CurveTransformKind.PARALLEL_SHIFT, amount=_to_decimal(amount))

    @classmethod
    def multiplicative(cls, factor: object) -> "CurveTransform":
        """Create a multiplicative scaling of the base zero rate."""
        return cls(kind=CurveTransformKind.MULTIPLICATIVE, amount=_to_decimal(factor))

    @classmethod
    def spread_overlay(cls, overlay: YieldCurve) -> "CurveTransform":
        """Create an additive overlay using the overlay curve's zero rate."""
        return cls(kind=CurveTransformKind.SPREAD_OVERLAY, overlay=overlay)

    @classmethod
    def function(cls, fn: Callable[[float, float], float]) -> "CurveTransform":
        """Create a custom tenor/base-zero transform."""
        return cls(kind=CurveTransformKind.FUNCTION, fn=fn)

    def apply(self, *, tenor: float, base_zero: float, date: Date) -> float:
        """Apply the transform to a continuously compounded base zero rate."""
        if self.kind is CurveTransformKind.PARALLEL_SHIFT:
            return base_zero + float(self.amount)
        if self.kind is CurveTransformKind.MULTIPLICATIVE:
            return base_zero * float(self.amount)
        if self.kind is CurveTransformKind.SPREAD_OVERLAY:
            if self.overlay is None:  # pragma: no cover - defensive
                return base_zero
            overlay_zero = self.overlay.zero_rate(date).convert_to(Compounding.CONTINUOUS).value()
            return base_zero + float(overlay_zero)
        if self.kind is CurveTransformKind.FUNCTION:
            if self.fn is None:  # pragma: no cover - defensive
                return base_zero
            return float(self.fn(float(tenor), float(base_zero)))
        return base_zero


@dataclass(frozen=True, slots=True)
class DerivedCurve(YieldCurve):
    """Curve derived from a base curve via a sequence of transforms."""

    base_curve: YieldCurve
    transforms: tuple[CurveTransform, ...]

    @classmethod
    def from_curve(cls, base_curve: YieldCurve, *transforms: CurveTransform) -> "DerivedCurve":
        """Construct a derived curve from a base curve and transforms."""
        return cls(base_curve=base_curve, transforms=tuple(transforms))

    @classmethod
    def parallel_shift(cls, base_curve: YieldCurve, amount: object) -> "DerivedCurve":
        """Construct a derived curve with a single parallel shift."""
        return cls.from_curve(base_curve, CurveTransform.parallel_shift(amount))

    @classmethod
    def spread_overlay(cls, base_curve: YieldCurve, overlay: YieldCurve) -> "DerivedCurve":
        """Construct a derived curve with a zero-rate spread overlay."""
        return cls.from_curve(base_curve, CurveTransform.spread_overlay(overlay))

    @classmethod
    def transformed(
        cls,
        base_curve: YieldCurve | RateCurve,
        fn: Callable[[float, float], float],
    ) -> "DerivedCurve":
        """Construct a derived curve backed by an arbitrary callable."""
        curve = base_curve if isinstance(base_curve, YieldCurve) else RateCurve(base_curve)
        return cls.from_curve(curve, CurveTransform.function(fn))

    def reference_date(self) -> Date:
        """Return the reference date of the base curve."""
        return self.base_curve.reference_date()

    def max_date(self) -> Date:
        """Return the maximum date supported by the base curve."""
        return self.base_curve.max_date()

    def zero_rate(self, date: Date) -> Yield:
        """Return the transformed continuously compounded zero rate."""
        tenor = max(float(self.reference_date().days_between(date)) / 365.0, 0.0)
        zero = float(self.base_curve.zero_rate(date).convert_to(Compounding.CONTINUOUS).value())
        for transform in self.transforms:
            zero = transform.apply(tenor=tenor, base_zero=zero, date=date)
        return Yield.new(_to_decimal(zero), Compounding.CONTINUOUS)

    def discount_factor(self, date: Date) -> Decimal:
        """Return the discount factor implied by the transformed zero rate."""
        tenor = max(float(self.reference_date().days_between(date)) / 365.0, 0.0)
        if tenor <= 0.0:
            return Decimal(1)
        zero = float(self.zero_rate(date).value())
        return _to_decimal(ValueConverter.zero_to_df(zero, tenor, Compounding.CONTINUOUS))


__all__ = ["CurveTransform", "CurveTransformKind", "DerivedCurve"]



"""Bond curve instruments and their market-convention helpers.

The curve-instrument layer provides the bond cashflow templates used when
building discount curves from government and other benchmark bonds.
"""

from __future__ import annotations

from .conventions import MarketConvention, day_count_factor
from .government import GovernmentCouponBond, GovernmentZeroCoupon

__all__ = [
    "GovernmentZeroCoupon",
    "GovernmentCouponBond",
    "MarketConvention",
    "day_count_factor",
]



from __future__ import annotations

from decimal import Decimal

from fuggers_py.bonds.types import CreditRating, Sector
from fuggers_py.core import Currency, Date
from fuggers_py.portfolio.types import Classification, Holding, MaturityBucket

from ._portfolio_helpers import make_holding


def test_holding_market_value() -> None:
    ref = Date.from_ymd(2024, 1, 1)
    holding = make_holding(
        ref, years=3, coupon="0.04", label="a", sector=Sector.CORPORATE, rating=CreditRating.BBB, quantity="1"
    )
    assert holding.market_value_amount == Decimal("100")


def test_holding_dirty_market_value() -> None:
    ref = Date.from_ymd(2024, 1, 1)
    base = make_holding(
        ref, years=3, coupon="0.04", label="a", sector=Sector.CORPORATE, rating=CreditRating.BBB, quantity="1"
    )
    holding = Holding(
        instrument=base.instrument,
        quantity=base.quantity,
        clean_price=base.clean_price,
        accrued_interest=Decimal("0.5"),
        classification=base.classification,
        rating_info=base.rating_info,
        sector_info=base.sector_info,
    )
    assert holding.dirty_market_value == Decimal("100.5")


def test_holding_base_currency_value() -> None:
    ref = Date.from_ymd(2024, 1, 1)
    base = make_holding(
        ref, years=3, coupon="0.04", label="a", sector=Sector.CORPORATE, rating=CreditRating.BBB, quantity="1"
    )
    holding = Holding(
        instrument=base.instrument,
        quantity=base.quantity,
        clean_price=base.clean_price,
        classification=Classification(currency=Currency.EUR),
        fx_rate=Decimal("1.10"),
    )
    assert holding.base_currency_value == Decimal("110.0")


def test_maturity_bucket_assignment() -> None:
    bucket = MaturityBucket(label="2-5Y", start_years=2.0, end_years=5.0)
    assert bucket.contains(3.0)
    assert not bucket.contains(5.0)


def test_classification_defaults() -> None:
    classification = Classification()
    assert classification.sector is None
    assert classification.custom_fields == {}


def test_credit_rating_ordering() -> None:
    assert CreditRating.A.score() < CreditRating.BB.score()



"""Street-convention yield (`fuggers_py.analytics.yields.street`).

The returned yield is a raw decimal rate, matching the bond solver's internal
representation rather than a quoted percentage.
"""

from __future__ import annotations

from fuggers_py.bonds.types import YieldConvention

from .solver import YieldSolver


def street_convention_yield(
    dirty_price: float,
    cashflows: list[float],
    times: list[float],
    *,
    frequency: int,
    initial_guess: float | None = None,
) -> float:
    """Return the street-convention yield as a raw decimal rate."""

    solver = YieldSolver()
    result = solver.solve(
        dirty_price=float(dirty_price),
        cashflows=[float(cf) for cf in cashflows],
        times=[float(t) for t in times],
        frequency=int(frequency),
        convention=YieldConvention.STREET_CONVENTION,
        initial_guess=initial_guess,
    )
    return float(result.yield_value)


__all__ = ["street_convention_yield"]



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.19.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Curve Scenarios, Key Rates, and Multi-Curve
#
# This notebook builds a small discount/projection setup, applies standard bumps, and shows how those shocks feed downstream floating-rate risk interpretation.

# %%
from __future__ import annotations

from decimal import Decimal

import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt
import pandas as pd

from fuggers_py.bonds import Tenor
from fuggers_py.core import Currency, Date
from fuggers_py.curves import KeyRateBump, MultiCurveEnvironmentBuilder, ParallelBump, RateIndex, ZeroCurveBuilder, key_rate_profile

from examples._shared.plotting import apply_research_style, percent_axis
from examples._shared.tables import display_table

apply_research_style()

# %%
reference_date = Date.from_ymd(2026, 1, 2)
discount_curve = (
    ZeroCurveBuilder(reference_date=reference_date)
    .add_rate(reference_date.add_years(1), Decimal('0.0405'))
    .add_rate(reference_date.add_years(5), Decimal('0.0385'))
    .add_rate(reference_date.add_years(10), Decimal('0.0390'))
    .build()
)
projection_curve = (
    ZeroCurveBuilder(reference_date=reference_date)
    .add_rate(reference_date.add_years(1), Decimal('0.0420'))
    .add_rate(reference_date.add_years(5), Decimal('0.0402'))
    .add_rate(reference_date.add_years(10), Decimal('0.0408'))
    .build()
)
sofr = RateIndex.new('SOFR', Tenor.parse('3M'), Currency.USD)
environment = (
    MultiCurveEnvironmentBuilder()
    .add_discount_curve(Currency.USD, discount_curve)
    .add_projection_curve(sofr, projection_curve)
    .build()
)

parallel_up = ParallelBump(0.0010).apply(discount_curve)
key_rate_5y = KeyRateBump([Tenor.parse('2Y'), Tenor.parse('5Y'), Tenor.parse('10Y')], Tenor.parse('5Y'), 0.0010).apply(discount_curve)

scenario_grid = [1.0, 2.0, 5.0, 7.0, 10.0]
scenario_table = pd.DataFrame(
    {
        'tenor_years': scenario_grid,
        'discount_base': [float(discount_curve.zero_rate(reference_date.add_days(int(365 * tenor))).value()) for tenor in scenario_grid],
        'discount_parallel_up': [float(parallel_up.zero_rate(reference_date.add_days(int(365 * tenor))).value()) for tenor in scenario_grid],
        'discount_key_rate_5y': [float(key_rate_5y.zero_rate(reference_date.add_days(int(365 * tenor))).value()) for tenor in scenario_grid],
        'projection_curve': [float(environment.projection_curve(sofr).zero_rate(reference_date.add_days(int(365 * tenor))).value()) for tenor in scenario_grid],
    }
)

display_table(
    scenario_table,
    percent_columns=['discount_base', 'discount_parallel_up', 'discount_key_rate_5y', 'projection_curve'],
    float_columns=['tenor_years'],
)

# %% [markdown]
# Discounting and projection are separated because floating-rate instruments discount cashflows off the funding curve but project coupons off the index-specific curve. That separation is what makes multi-curve scenario analysis useful in practice.

# %%
figure, axes = plt.subplots(1, 2, figsize=(12, 4.6))
axes[0].plot(scenario_table['tenor_years'], 100 * scenario_table['discount_base'], label='Base discount', linewidth=2)
axes[0].plot(scenario_table['tenor_years'], 100 * scenario_table['discount_parallel_up'], label='Parallel +10bp', linewidth=2)
axes[0].plot(scenario_table['tenor_years'], 100 * scenario_table['discount_key_rate_5y'], label='5Y key-rate +10bp', linewidth=2)
axes[0].plot(scenario_table['tenor_years'], 100 * scenario_table['projection_curve'], label='Projection curve', linewidth=2)
axes[0].set_title('Base And Shocked Curves')
axes[0].set_xlabel('Tenor (years)')
axes[0].set_ylabel('Zero rate')
percent_axis(axes[0])
axes[0].legend(title='')

profile = key_rate_profile(Tenor.parse('5Y'))
profile_table = pd.DataFrame({'tenor': list(profile.keys()), 'shock': list(profile.values())})
axes[1].bar([str(tenor) for tenor in profile_table['tenor']], [10000 * float(value) for value in profile_table['shock']])
axes[1].set_title('Standard 5Y Key-Rate Bump Profile')
axes[1].set_xlabel('Key tenor')
axes[1].set_ylabel('Shock (bps)')
plt.tight_layout()
figure



from __future__ import annotations

import pytest

from fuggers_py.core import Currency
from fuggers_py.data import (
    CurrencyPair,
    VolSurfaceId,
    YearMonth,
)


def test_currency_pair_normalization_and_roundtrip() -> None:
    pair = CurrencyPair.parse("usd eur")

    assert pair == CurrencyPair.from_string("USDEUR")
    assert pair.base is Currency.USD
    assert pair.quote is Currency.EUR
    assert pair.as_str() == "USD/EUR"
    assert {pair: "fx"}[CurrencyPair.parse("usd/eur")] == "fx"


def test_year_month_roundtrip_and_validation() -> None:
    year_month = YearMonth.parse("2026-03")

    assert year_month == YearMonth.from_string("2026-03")
    assert year_month.year == 2026
    assert year_month.month == 3
    assert year_month.as_str() == "2026-03"
    assert VolSurfaceId.from_string("usd.swaption").as_str() == "usd.swaption"

    with pytest.raises(ValueError):
        YearMonth.parse("2026-13")




from __future__ import annotations

from decimal import Decimal

import numpy as np
import pytest

from fuggers_py.bonds.types import Tenor
from fuggers_py.core import Compounding, Date
from fuggers_py.curves.calibration.global_fit import FitterConfig, GlobalFitter
from fuggers_py.curves.calibration.instruments import Deposit, InstrumentSet
from fuggers_py.curves.conversion import ValueConverter
from fuggers_py.curves.curves.discrete import DiscreteCurve, ExtrapolationMethod, InterpolationMethod
from fuggers_py.curves.errors import InvalidCurveInput, UnsupportedValueType
from fuggers_py.curves.value_type import ValueType
from fuggers_py.curves.wrappers import CreditCurve
from fuggers_py.math.optimization import OptimizationConfig


def test_value_converter_validates_forward_inputs_and_risky_discount_factor_formula() -> None:
    with pytest.raises(InvalidCurveInput):
        ValueConverter.forward_rate_from_dfs(0.99, 0.98, 2.0, 2.0, Compounding.CONTINUOUS)

    risky_df = ValueConverter.risky_discount_factor(0.95, 0.90, 0.40)

    assert risky_df == pytest.approx(0.95 * (0.90 + (1.0 - 0.90) * 0.40), abs=1e-12)


def test_credit_curve_rejects_invalid_recovery_survival_and_value_types() -> None:
    ref = Date.from_ymd(2024, 1, 1)
    survival_curve = DiscreteCurve(
        ref,
        tenors=[1.0, 5.0],
        values=[1.05, 1.02],
        value_type=ValueType.survival_probability(),
        interpolation_method=InterpolationMethod.LINEAR,
        extrapolation_method=ExtrapolationMethod.FLAT,
    )
    unsupported_curve = DiscreteCurve(
        ref,
        tenors=[1.0, 5.0],
        values=[0.03, 0.035],
        value_type=ValueType.zero_rate(Compounding.CONTINUOUS),
        interpolation_method=InterpolationMethod.LINEAR,
        extrapolation_method=ExtrapolationMethod.FLAT,
    )

    with pytest.raises(InvalidCurveInput):
        CreditCurve(unsupported_curve, recovery_rate=Decimal("1.0"))

    with pytest.raises(InvalidCurveInput):
        CreditCurve(survival_curve).survival_probability_at_tenor(2.0)

    with pytest.raises(UnsupportedValueType):
        CreditCurve(unsupported_curve).hazard_rate_at_tenor(2.0)


def test_global_fitter_validates_inputs_and_initial_parameter_length() -> None:
    ref = Date.from_ymd(2024, 1, 1)
    fitter = GlobalFitter(ref)

    with pytest.raises(InvalidCurveInput):
        fitter.fit_zero_rates(np.array([[1.0, 2.0]]), np.array([0.03, 0.04]))

    with pytest.raises(InvalidCurveInput):
        fitter.fit_zero_rates(np.array([1.0, 2.0]), np.array([0.03]))

    with pytest.raises(InvalidCurveInput):
        fitter.fit_zero_rates(np.array([0.0, 2.0]), np.array([0.03, 0.04]))

    with pytest.raises(InvalidCurveInput):
        fitter.fit_zero_rates(np.array([1.0, np.nan]), np.array([0.03, 0.04]))

    bad_initial = FitterConfig(initial_parameters=np.array([0.03, 0.04]))
    with pytest.raises(InvalidCurveInput):
        GlobalFitter(ref, config=bad_initial).fit_zero_rates(np.array([1.0, 3.0]), np.array([0.03, 0.035]))


def test_global_fitter_can_fit_simple_deposit_instruments_close_to_quotes() -> None:
    ref = Date.from_ymd(2024, 1, 1)
    instruments = InstrumentSet(
        [
            Deposit(ref, Decimal("0.03"), tenor=Tenor.parse("1Y")),
            Deposit(ref, Decimal("0.032"), tenor=Tenor.parse("3Y")),
            Deposit(ref, Decimal("0.035"), tenor=Tenor.parse("5Y")),
            Deposit(ref, Decimal("0.038"), tenor=Tenor.parse("10Y")),
        ]
    )
    fitter = GlobalFitter(
        ref,
        config=FitterConfig(optimization=OptimizationConfig(max_iterations=200, tolerance=1e-12)),
    )

    result = fitter.fit_instruments(instruments)

    assert result.parameters.shape == (4,)
    assert result.curve.zero_rate_at_tenor(1.0) > 0.0
    for instrument in instruments.instruments:
        assert float(instrument.par_rate(result.curve)) == pytest.approx(float(instrument.quote), abs=5e-4)



"""Contribution analytics.

The contribution surface provides holding-level and bucket-level attribution
for duration, DV01, spread, and return decomposition. Metrics use the same
portfolio weights as the main analytics layer.
"""

from __future__ import annotations

from dataclasses import dataclass

from ..portfolio import Portfolio
from .attribution import (
    AggregatedAttribution,
    AttributionInput,
    BucketContribution,
    HoldingAttribution,
    PortfolioAttribution,
    SectorAttribution,
    aggregated_attribution,
    attribution_summary,
    calculate_attribution,
    duration_difference_by_sector,
    estimate_income_returns,
    estimate_rate_returns,
    estimate_spread_returns,
    spread_difference_by_sector,
    top_contributors,
    weights_sum_check,
)
from .risk import (
    Cs01Contributions,
    DurationContributions,
    Dv01Contributions,
    HoldingContribution,
    SpreadContributions,
    contribution_by_rating,
    contribution_by_sector,
    cs01_contributions,
    duration_contributions,
    dv01_contributions,
    spread_contributions,
)


@dataclass(frozen=True, slots=True)
class Contribution:
    """Convenience wrapper for portfolio contribution helpers."""

    portfolio: Portfolio

    def by_position(self, curve, settlement_date):
        """Return holding-level attribution for the portfolio."""

        return attribution_summary(self.portfolio, curve=curve, settlement_date=settlement_date)

    def aggregate(self, curve, settlement_date, *, assumptions: AttributionInput | None = None):
        """Return aggregated contribution and return decomposition."""

        return aggregated_attribution(
            self.portfolio,
            curve=curve,
            settlement_date=settlement_date,
            assumptions=assumptions,
        )


__all__ = [
    "AggregatedAttribution",
    "AttributionInput",
    "BucketContribution",
    "Cs01Contributions",
    "Contribution",
    "DurationContributions",
    "Dv01Contributions",
    "HoldingAttribution",
    "HoldingContribution",
    "PortfolioAttribution",
    "SectorAttribution",
    "SpreadContributions",
    "aggregated_attribution",
    "attribution_summary",
    "calculate_attribution",
    "contribution_by_rating",
    "contribution_by_sector",
    "cs01_contributions",
    "duration_difference_by_sector",
    "duration_contributions",
    "dv01_contributions",
    "estimate_income_returns",
    "estimate_rate_returns",
    "estimate_spread_returns",
    "spread_contributions",
    "spread_difference_by_sector",
    "top_contributors",
    "weights_sum_check",
]



from __future__ import annotations

from decimal import Decimal

import pytest

from fuggers_py.bonds.instruments import FixedBondBuilder
from fuggers_py.core import Compounding, Date, Frequency, Yield


def test_fixed_bond_accrued_interest_basic() -> None:
    bond = (
        FixedBondBuilder.new()
        .with_issue_date(Date.from_ymd(2024, 1, 1))
        .with_maturity_date(Date.from_ymd(2025, 1, 1))
        .with_coupon_rate(Decimal("0.06"))
        .with_frequency(Frequency.SEMI_ANNUAL)
        .build()
    )
    # Half way through the first coupon period: accrued should be half the coupon.
    accrued = bond.accrued_interest(Date.from_ymd(2024, 4, 1))
    assert accrued == Decimal("1.5")

    assert bond.accrued_interest(Date.from_ymd(2024, 1, 1)) == Decimal(0)
    assert bond.accrued_interest(Date.from_ymd(2024, 7, 1)) == Decimal(0)


def test_price_yield_round_trip_simple_fixed_bond() -> None:
    bond = (
        FixedBondBuilder.new()
        .with_issue_date(Date.from_ymd(2024, 1, 1))
        .with_maturity_date(Date.from_ymd(2026, 1, 1))
        .with_coupon_rate(Decimal("0.05"))
        .with_frequency(Frequency.SEMI_ANNUAL)
        .build()
    )

    settlement = Date.from_ymd(2024, 1, 1)
    ytm = Yield.new(Decimal("0.05"), Compounding.SEMI_ANNUAL)

    price = bond.price_from_yield(ytm, settlement)
    ytm2 = bond.yield_from_price(price.clean, settlement)

    assert float(price.clean.as_percentage()) == pytest.approx(100.0, abs=1e-10)
    assert float(ytm2.ytm.value()) == pytest.approx(0.05, abs=1e-12)



"""Macaulay duration (`fuggers_py.analytics.risk.duration.macaulay`).

The returned value is a positive magnitude in year units.
"""

from __future__ import annotations

from decimal import Decimal

from fuggers_py.bonds.risk import RiskMetrics
from fuggers_py.bonds.traits import Bond
from fuggers_py.core.types import Date, Yield


def macaulay_duration(bond: Bond, ytm: Yield, settlement_date: Date) -> Decimal:
    """Return Macaulay duration for the supplied bond."""

    metrics = RiskMetrics.from_bond(bond, ytm, settlement_date)
    return metrics.macaulay_duration


__all__ = ["macaulay_duration"]



from __future__ import annotations

from decimal import Decimal

import pytest

import fuggers_py.bonds.pricing.yield_engine as yield_engine_module

from fuggers_py.analytics.cashflows.irregular import IrregularPeriodHandler
from fuggers_py.analytics.error import AnalyticsError
from fuggers_py.analytics.yields.street import street_convention_yield
from fuggers_py.bonds.instruments import FixedBondBuilder
from fuggers_py.bonds.pricing.yield_engine import StandardYieldEngine, _prepare_cashflows
from fuggers_py.bonds.types import YieldCalculationRules
from fuggers_py.core import Currency, Date, Frequency
from fuggers_py.core.daycounts import ActActIcma, DayCountConvention
from fuggers_py.math.errors import ConvergenceFailed


def _treasury_bond():
    return (
        FixedBondBuilder.new()
        .with_issue_date(Date.from_ymd(2024, 1, 1))
        .with_maturity_date(Date.from_ymd(2025, 7, 1))
        .with_coupon_rate(Decimal("0.04"))
        .with_frequency(Frequency.SEMI_ANNUAL)
        .with_currency(Currency.USD)
        .with_rules(YieldCalculationRules.us_treasury())
        .build()
    )


def test_prepare_cashflows_accumulates_partial_icma_periods_from_settlement() -> None:
    bond = _treasury_bond()
    settlement = Date.from_ymd(2024, 4, 1)
    prepared = _prepare_cashflows(
        bond.cash_flows(),
        settlement_date=settlement,
        rules=bond.rules(),
    )

    icma = ActActIcma.semi_annual()
    first = bond.cash_flows()[0]
    assert first.accrual_start is not None
    assert first.accrual_end is not None
    first_period = icma.year_fraction_with_period(
        settlement,
        first.accrual_end,
        first.accrual_start,
        first.accrual_end,
    )

    assert prepared[0].years == pytest.approx(float(first_period), abs=1e-12)
    assert prepared[1].years == pytest.approx(prepared[0].years + 0.5, abs=1e-12)
    assert prepared[2].years == pytest.approx(prepared[1].years + 0.5, abs=1e-12)


def test_standard_yield_engine_dirty_price_is_monotone_in_yield() -> None:
    bond = _treasury_bond()
    engine = StandardYieldEngine()
    settlement = Date.from_ymd(2024, 4, 1)

    low = engine.dirty_price_from_yield(bond.cash_flows(), yield_rate=0.03, settlement_date=settlement, rules=bond.rules())
    mid = engine.dirty_price_from_yield(bond.cash_flows(), yield_rate=0.05, settlement_date=settlement, rules=bond.rules())
    high = engine.dirty_price_from_yield(bond.cash_flows(), yield_rate=0.08, settlement_date=settlement, rules=bond.rules())

    assert low > mid > high > 0.0


def test_standard_yield_engine_falls_back_to_brent_when_newton_fails(monkeypatch: pytest.MonkeyPatch) -> None:
    bond = _treasury_bond()
    settlement = Date.from_ymd(2024, 4, 1)
    pricing_engine = StandardYieldEngine()
    dirty = Decimal(
        str(
            pricing_engine.dirty_price_from_yield(
                bond.cash_flows(),
                yield_rate=0.05,
                settlement_date=settlement,
                rules=bond.rules(),
            )
        )
    )
    clean = dirty - bond.accrued_interest(settlement)

    def _fail_newton(*_args, **_kwargs):
        raise ConvergenceFailed(iterations=1, residual=1.0)

    monkeypatch.setattr(yield_engine_module, "newton_raphson", _fail_newton)

    result = StandardYieldEngine().yield_from_price(
        bond.cash_flows(),
        clean_price=clean,
        accrued=bond.accrued_interest(settlement),
        settlement_date=settlement,
        rules=bond.rules(),
    )

    assert result.method == "Brent"
    assert result.yield_rate == pytest.approx(0.05, abs=1e-10)


def test_standard_yield_engine_surfaces_failure_after_all_solver_paths_exhausted(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    bond = _treasury_bond()
    settlement = Date.from_ymd(2024, 4, 1)

    def _fail_newton(*_args, **_kwargs):
        raise ConvergenceFailed(iterations=1, residual=1.0)

    def _fail_brent(*_args, **_kwargs):
        raise ConvergenceFailed(iterations=2, residual=1.0)

    monkeypatch.setattr(yield_engine_module, "newton_raphson", _fail_newton)
    monkeypatch.setattr(yield_engine_module, "brent", _fail_brent)

    with pytest.raises(yield_engine_module.YieldConvergenceFailed):
        StandardYieldEngine().yield_from_price(
            bond.cash_flows(),
            clean_price=Decimal("99.50"),
            accrued=bond.accrued_interest(settlement),
            settlement_date=settlement,
            rules=bond.rules(),
        )


def test_street_convention_yield_recovers_known_solution_and_moves_inverse_to_price() -> None:
    cashflows = [2.5, 102.5]
    times = [0.5, 1.0]
    expected_yield = 0.045
    dirty_price = sum(
        cf / ((1.0 + expected_yield / 2.0) ** (t * 2.0))
        for cf, t in zip(cashflows, times, strict=True)
    )

    solved = street_convention_yield(
        dirty_price,
        cashflows,
        times,
        frequency=2,
        initial_guess=-0.25,
    )
    richer = street_convention_yield(dirty_price + 0.5, cashflows, times, frequency=2)

    assert solved == pytest.approx(expected_yield, abs=1e-12)
    assert richer < solved


def test_irregular_period_handler_uses_day_count_objects_consistently() -> None:
    handler = IrregularPeriodHandler.new(DayCountConvention.ACT_365_FIXED)
    start = Date.from_ymd(2024, 1, 1)
    end = Date.from_ymd(2024, 2, 1)

    assert handler.days_between(start, end) == 31
    assert handler.year_fraction(start, end) == pytest.approx(float(Decimal(31) / Decimal(365)), abs=1e-12)
    assert handler.annual_factor(start, end) == pytest.approx(float(Decimal(365) / Decimal(31)), abs=1e-12)

    with pytest.raises(AnalyticsError, match="must be positive"):
        handler.annual_factor(start, start)



"""Capital-charge spread overlays.

Capital overlays return raw decimal spread adjustments derived from capital
consumption, hurdle rate, and pass-through assumptions.
"""

from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal

from .balance_sheet import BaseSpreadAdjustment, SpreadAdjustmentBreakdown


def _to_decimal(value: object) -> Decimal:
    if isinstance(value, Decimal):
        return value
    return Decimal(str(value))


@dataclass(frozen=True, slots=True, kw_only=True)
class CapitalAdjustmentBreakdown(SpreadAdjustmentBreakdown):
    """Detailed capital spread adjustment breakdown."""

    exposure: Decimal
    risk_weight: Decimal
    capital_ratio: Decimal
    hurdle_rate: Decimal
    pass_through: Decimal
    capital_consumed: Decimal
    annual_capital_cost: Decimal
    passed_through_cost: Decimal

    def __post_init__(self) -> None:
        SpreadAdjustmentBreakdown.__post_init__(self)
        object.__setattr__(self, "exposure", _to_decimal(self.exposure))
        object.__setattr__(self, "risk_weight", _to_decimal(self.risk_weight))
        object.__setattr__(self, "capital_ratio", _to_decimal(self.capital_ratio))
        object.__setattr__(self, "hurdle_rate", _to_decimal(self.hurdle_rate))
        object.__setattr__(self, "pass_through", _to_decimal(self.pass_through))
        object.__setattr__(self, "capital_consumed", _to_decimal(self.capital_consumed))
        object.__setattr__(self, "annual_capital_cost", _to_decimal(self.annual_capital_cost))
        object.__setattr__(self, "passed_through_cost", _to_decimal(self.passed_through_cost))


def capital_adjustment_breakdown(
    *,
    exposure: object,
    risk_weight: object,
    capital_ratio: object,
    hurdle_rate: object,
    pass_through: object = Decimal(1),
    name: str = "capital",
) -> CapitalAdjustmentBreakdown:
    """Return the capital spread adjustment breakdown as raw decimals."""
    resolved_exposure = _to_decimal(exposure)
    resolved_risk_weight = _to_decimal(risk_weight)
    resolved_capital_ratio = _to_decimal(capital_ratio)
    resolved_hurdle_rate = _to_decimal(hurdle_rate)
    resolved_pass_through = _to_decimal(pass_through)
    if resolved_exposure <= Decimal(0):
        raise ValueError("Capital spread adjustment requires positive exposure.")
    if resolved_risk_weight < Decimal(0):
        raise ValueError("Capital spread adjustment requires non-negative risk_weight.")
    if resolved_capital_ratio < Decimal(0) or resolved_capital_ratio > Decimal(1):
        raise ValueError("Capital spread adjustment requires capital_ratio between 0 and 1.")
    if resolved_hurdle_rate < Decimal(0):
        raise ValueError("Capital spread adjustment requires non-negative hurdle_rate.")
    if resolved_pass_through < Decimal(0):
        raise ValueError("Capital spread adjustment requires non-negative pass_through.")

    capital_consumed = resolved_exposure * resolved_risk_weight * resolved_capital_ratio
    annual_capital_cost = capital_consumed * resolved_hurdle_rate
    passed_through_cost = annual_capital_cost * resolved_pass_through
    spread_adjustment = passed_through_cost / resolved_exposure
    return CapitalAdjustmentBreakdown(
        name=name,
        spread_adjustment=spread_adjustment,
        description="Capital consumption translated into an annualized spread overlay.",
        exposure=resolved_exposure,
        risk_weight=resolved_risk_weight,
        capital_ratio=resolved_capital_ratio,
        hurdle_rate=resolved_hurdle_rate,
        pass_through=resolved_pass_through,
        capital_consumed=capital_consumed,
        annual_capital_cost=annual_capital_cost,
        passed_through_cost=passed_through_cost,
    )


def capital_spread_adjustment(
    *,
    exposure: object,
    risk_weight: object,
    capital_ratio: object,
    hurdle_rate: object,
    pass_through: object = Decimal(1),
) -> Decimal:
    """Return the capital spread adjustment as a raw decimal."""
    return capital_adjustment_breakdown(
        exposure=exposure,
        risk_weight=risk_weight,
        capital_ratio=capital_ratio,
        hurdle_rate=hurdle_rate,
        pass_through=pass_through,
    ).spread_adjustment


@dataclass(frozen=True, slots=True)
class CapitalSpreadAdjustment(BaseSpreadAdjustment):
    """Capital-charge spread adjustment with stored input assumptions."""

    exposure: Decimal
    risk_weight: Decimal
    capital_ratio: Decimal
    hurdle_rate: Decimal
    pass_through: Decimal = Decimal(1)
    name: str = "capital"

    def __post_init__(self) -> None:
        object.__setattr__(self, "exposure", _to_decimal(self.exposure))
        object.__setattr__(self, "risk_weight", _to_decimal(self.risk_weight))
        object.__setattr__(self, "capital_ratio", _to_decimal(self.capital_ratio))
        object.__setattr__(self, "hurdle_rate", _to_decimal(self.hurdle_rate))
        object.__setattr__(self, "pass_through", _to_decimal(self.pass_through))
        object.__setattr__(self, "name", self.name.strip())

    def breakdown(self) -> CapitalAdjustmentBreakdown:
        """Return the capital adjustment breakdown."""
        return capital_adjustment_breakdown(
            exposure=self.exposure,
            risk_weight=self.risk_weight,
            capital_ratio=self.capital_ratio,
            hurdle_rate=self.hurdle_rate,
            pass_through=self.pass_through,
            name=self.name,
        )


__all__ = [
    "CapitalAdjustmentBreakdown",
    "CapitalSpreadAdjustment",
    "capital_adjustment_breakdown",
    "capital_spread_adjustment",
]



from __future__ import annotations

from decimal import Decimal

from fuggers_py.analytics.spreads import ParParAssetSwap, ProceedsAssetSwap
from fuggers_py.bonds import FixedBondBuilder, YieldCalculationRules
from fuggers_py.bonds.types import ASWType
from fuggers_py.core import Currency, Date, Frequency
from fuggers_py.data import AnalyticsCurves
from fuggers_py.rates import AssetSwap, AssetSwapPricer, FloatingLegSpec, PayReceive, ScheduleDefinition

from ._rates_helpers import flat_curve, multicurve_analytics_curves, rate_index


def _bond(settlement: Date):
    return (
        FixedBondBuilder.new()
        .with_issue_date(Date.from_ymd(2021, 1, 15))
        .with_maturity_date(settlement.add_years(5))
        .with_coupon_rate(Decimal("0.0450"))
        .with_frequency(Frequency.SEMI_ANNUAL)
        .with_currency(Currency.USD)
        .with_rules(YieldCalculationRules.us_corporate())
        .build()
    )


def _curves(settlement: Date) -> tuple[AnalyticsCurves, object]:
    discount = flat_curve(settlement, Decimal("0.0330"))
    overnight = flat_curve(settlement, Decimal("0.0340"))
    term = flat_curve(settlement, Decimal("0.0360"))
    projection_index = rate_index("TERM", "6M")
    base_curves = multicurve_analytics_curves(
        discount_curve=discount,
        discount_currency=Currency.USD,
        forward_curve=overnight,
        projection_curves={projection_index: term},
    )
    curves = AnalyticsCurves(
        discount_curve=base_curves.discount_curve,
        forward_curve=base_curves.forward_curve,
        repo_curve=flat_curve(settlement, Decimal("0.0310")),
        collateral_curve=flat_curve(settlement, Decimal("0.0320")),
        multicurve_environment=base_curves.multicurve_environment,
        projection_curves=base_curves.projection_curves,
    )
    return curves, term


def test_asset_swap_par_spread_matches_low_level_kernel_for_par_par() -> None:
    settlement = Date.from_ymd(2026, 1, 15)
    bond = _bond(settlement)
    curves, term_curve = _curves(settlement)
    asset_swap = AssetSwap(
        bond=bond,
        settlement_date=settlement,
        floating_leg=FloatingLegSpec(
            pay_receive=PayReceive.RECEIVE,
            notional=Decimal("100"),
            index_name="TERM",
            index_tenor="6M",
            currency=Currency.USD,
            schedule=ScheduleDefinition(frequency=Frequency.QUARTERLY),
        ),
        market_dirty_price=Decimal("101.25"),
    )

    expected = ParParAssetSwap(term_curve).calculate(bond, Decimal("101.25"), settlement)
    result = AssetSwapPricer().par_spread(asset_swap, curves)

    assert result == expected


def test_asset_swap_pv_is_zero_when_the_trade_spread_equals_proceeds_par_spread() -> None:
    settlement = Date.from_ymd(2026, 1, 15)
    bond = _bond(settlement)
    curves, term_curve = _curves(settlement)
    par_spread = ProceedsAssetSwap(term_curve).calculate(bond, Decimal("101.25"), settlement)
    asset_swap = AssetSwap(
        bond=bond,
        settlement_date=settlement,
        floating_leg=FloatingLegSpec(
            pay_receive=PayReceive.RECEIVE,
            notional=Decimal("100"),
            index_name="TERM",
            index_tenor="6M",
            currency=Currency.USD,
            schedule=ScheduleDefinition(frequency=Frequency.QUARTERLY),
        ),
        asset_swap_type=ASWType.PROCEEDS,
        market_dirty_price=Decimal("101.25"),
        quoted_spread=par_spread,
    )

    assert AssetSwapPricer().pv(asset_swap, curves) == Decimal(0)



"""Flat-forward interpolation for zero rates with piecewise-constant forwards."""

from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from numpy.typing import ArrayLike, NDArray

from ..errors import ExtrapolationNotAllowed, InsufficientData, InvalidInput
from ..utils import assert_finite, assert_finite_array, assert_same_length, assert_strictly_increasing, bisect_segment


@dataclass(frozen=True, slots=True)
class FlatForward:
    """Interpolate zero rates while keeping forward rates constant by segment.

    The input curve is interpreted as zero rates at strictly increasing positive
    tenors. Between pillars the implied instantaneous forward rate is held
    constant; extrapolation beyond the last tenor reuses the last forward rate
    when enabled.
    """

    _tenors: NDArray[np.float64]
    _zero_rates: NDArray[np.float64]
    _forwards: NDArray[np.float64]
    _allow_extrapolation: bool = False

    def __init__(self, tenors: ArrayLike, zero_rates: ArrayLike, allow_extrapolation: bool = False) -> None:
        """Validate the tenor grid and precompute piecewise-constant forwards."""

        t = assert_strictly_increasing(tenors, name="tenors")
        r = assert_finite_array(zero_rates, name="zero_rates").astype(float, copy=False)
        assert_same_length(t, r, a_name="tenors", b_name="zero_rates")
        if t.size < 2:
            raise InsufficientData(required=2, actual=int(t.size))
        if np.any(t <= 0.0):
            raise InvalidInput("tenors must be strictly positive.")

        forwards = np.empty(int(t.size - 1), dtype=float)
        for i in range(forwards.size):
            t0 = float(t[i])
            t1 = float(t[i + 1])
            r0 = float(r[i])
            r1 = float(r[i + 1])
            forwards[i] = (r1 * t1 - r0 * t0) / (t1 - t0)

        object.__setattr__(self, "_tenors", t)
        object.__setattr__(self, "_zero_rates", r)
        object.__setattr__(self, "_forwards", forwards)
        object.__setattr__(self, "_allow_extrapolation", bool(allow_extrapolation))

    @classmethod
    def new(cls, tenors: ArrayLike, zero_rates: ArrayLike) -> "FlatForward":
        """Construct a flat-forward interpolator from tenors and zero rates."""

        return cls(tenors, zero_rates)

    @property
    def allows_extrapolation(self) -> bool:
        return bool(self._allow_extrapolation)

    def with_extrapolation(self) -> "FlatForward":
        """Return a copy that allows evaluation outside the tenor range."""

        return FlatForward(self._tenors, self._zero_rates, allow_extrapolation=True)

    def tenors(self) -> NDArray[np.float64]:
        """Return a copy of the tenor grid."""

        return self._tenors.copy()

    def zero_rates(self) -> NDArray[np.float64]:
        """Return a copy of the input zero rates."""

        return self._zero_rates.copy()

    def forward_rates_vec(self) -> NDArray[np.float64]:
        """Return the cached piecewise-constant forward rates."""

        return self._forwards.copy()

    def min_x(self) -> float:
        return float(self._tenors[0])

    def max_x(self) -> float:
        return float(self._tenors[-1])

    def in_range(self, x: float) -> bool:
        x = float(x)
        return self.min_x() <= x <= self.max_x()

    def _validate_t(self, t: float) -> float:
        t = assert_finite(t, name="t")
        if t <= 0.0:
            raise InvalidInput("t must be positive.")
        if not self.allows_extrapolation and not self.in_range(t):
            raise ExtrapolationNotAllowed(x=t, min=self.min_x(), max=self.max_x())
        return t

    def forward_rate(self, t: float) -> float:
        """Return the instantaneous forward rate implied at tenor ``t``."""

        t = self._validate_t(t)
        if t < self._tenors[0]:
            return float(self._zero_rates[0])
        if t > self._tenors[-1]:
            return float(self._forwards[-1])
        i = bisect_segment(self._tenors, t)
        return float(self._forwards[i])

    def interpolate(self, t: float) -> float:
        """Return the zero rate implied at tenor ``t``."""

        t = self._validate_t(t)
        if t <= self._tenors[0]:
            return float(self._zero_rates[0])
        if t >= self._tenors[-1]:
            if t == self._tenors[-1]:
                return float(self._zero_rates[-1])
            f_last = float(self._forwards[-1])
            r_last = float(self._zero_rates[-1])
            t_last = float(self._tenors[-1])
            return (r_last * t_last + f_last * (t - t_last)) / t

        i = bisect_segment(self._tenors, t)
        t0 = float(self._tenors[i])
        r0 = float(self._zero_rates[i])
        f = float(self._forwards[i])
        return (r0 * t0 + f * (t - t0)) / t

    def derivative(self, t: float) -> float:
        """Return the derivative of the zero-rate curve at tenor ``t``."""

        t = self._validate_t(t)
        if t <= self._tenors[0]:
            return 0.0

        if t >= self._tenors[-1]:
            if t == self._tenors[-1]:
                i = int(self._tenors.size - 2)
            else:
                i = int(self._tenors.size - 2)
        else:
            i = bisect_segment(self._tenors, t)

        t0 = float(self._tenors[i])
        r0 = float(self._zero_rates[i])
        f = float(self._forwards[i])
        return -(t0 * (r0 - f)) / (t**2)



"""Stress impact helpers."""

from __future__ import annotations

from decimal import Decimal

from ..analytics.base import PortfolioAnalytics
from ..portfolio import Portfolio
from ..types import StressResult
from .scenarios import KeyRateShiftScenario, RateShockScenario, SpreadShockScenario, StressSummary, summarize_results


def _run_stress_result(portfolio: Portfolio, *, curve, settlement_date, scenario: object) -> tuple[str, StressResult] | None:
    if isinstance(scenario, RateShockScenario):
        return (
            scenario.name,
            rate_shock_impact(
                portfolio,
                curve=curve,
                settlement_date=settlement_date,
                bump_bps=scenario.bump_bps,
                scenario_name=scenario.name,
            ),
        )
    if isinstance(scenario, SpreadShockScenario):
        return (
            scenario.name,
            spread_shock_result(
                portfolio,
                curve=curve,
                settlement_date=settlement_date,
                bump_bps=scenario.bump_bps,
                scenario_name=scenario.name,
            ),
        )
    if isinstance(scenario, KeyRateShiftScenario):
        return (
            scenario.name,
            key_rate_shift_result(
                portfolio,
                curve=curve,
                settlement_date=settlement_date,
                tenor_shocks_bps=scenario.tenor_shocks_bps,
                scenario_name=scenario.name,
            ),
        )
    return None


def rate_shock_impact(
    portfolio: Portfolio,
    *,
    curve,
    settlement_date,
    bump_bps: Decimal,
    scenario_name: str | None = None,
) -> StressResult:
    """Return the dirty-PV change for a parallel rate shock in bps."""

    metrics = PortfolioAnalytics(portfolio).metrics(curve, settlement_date)
    positions = PortfolioAnalytics(portfolio).position_metrics(curve, settlement_date)
    actual_change = -(metrics.dv01 * bump_bps)
    return StressResult(
        base_dirty_pv=metrics.dirty_pv,
        stressed_dirty_pv=metrics.dirty_pv + actual_change,
        actual_change=actual_change,
        dv01_approximation=actual_change,
        scenario_name=scenario_name or f"parallel_{bump_bps}",
        breakdown={item.name: -(item.dv01 * bump_bps) for item in positions},
    )


def parallel_shift_impact(
    portfolio: Portfolio,
    *,
    curve,
    settlement_date,
    bump_bps: Decimal,
    scenario_name: str | None = None,
) -> StressResult:
    """Compatibility alias for :func:`rate_shock_impact`."""

    return rate_shock_impact(
        portfolio,
        curve=curve,
        settlement_date=settlement_date,
        bump_bps=bump_bps,
        scenario_name=scenario_name,
    )


def spread_shock_impact(portfolio: Portfolio, *, curve, settlement_date, bump_bps: Decimal) -> Decimal:
    """Return the dirty-PV change for a parallel spread shock in bps."""

    metrics = PortfolioAnalytics(portfolio).metrics(curve, settlement_date)
    return -(metrics.cs01 * bump_bps)


def key_rate_shift_impact(portfolio: Portfolio, *, curve, settlement_date, tenor_shocks_bps: dict[str, Decimal]) -> Decimal:
    """Return the dirty-PV change for tenor-specific key-rate shocks."""

    profile = PortfolioAnalytics(portfolio).metrics(curve, settlement_date).key_rate_profile
    return sum((profile.get(tenor, Decimal(0)) * shock for tenor, shock in tenor_shocks_bps.items()), Decimal(0))


def spread_shock_result(
    portfolio: Portfolio,
    *,
    curve,
    settlement_date,
    bump_bps: Decimal,
    scenario_name: str | None = None,
) -> StressResult:
    """Return a typed result for a parallel spread shock."""

    metrics = PortfolioAnalytics(portfolio).metrics(curve, settlement_date)
    change = spread_shock_impact(portfolio, curve=curve, settlement_date=settlement_date, bump_bps=bump_bps)
    positions = PortfolioAnalytics(portfolio).position_metrics(curve, settlement_date)
    return StressResult(
        base_dirty_pv=metrics.dirty_pv,
        stressed_dirty_pv=metrics.dirty_pv + change,
        actual_change=change,
        dv01_approximation=Decimal(0),
        scenario_name=scenario_name or f"spread_{bump_bps}",
        breakdown={item.name: -((item.cs01 or Decimal(0)) * bump_bps) for item in positions},
    )


def key_rate_shift_result(
    portfolio: Portfolio,
    *,
    curve,
    settlement_date,
    tenor_shocks_bps: dict[str, Decimal],
    scenario_name: str | None = None,
) -> StressResult:
    """Return a typed result for a key-rate shift scenario."""

    metrics = PortfolioAnalytics(portfolio).metrics(curve, settlement_date)
    change = key_rate_shift_impact(
        portfolio,
        curve=curve,
        settlement_date=settlement_date,
        tenor_shocks_bps=tenor_shocks_bps,
    )
    return StressResult(
        base_dirty_pv=metrics.dirty_pv,
        stressed_dirty_pv=metrics.dirty_pv + change,
        actual_change=change,
        dv01_approximation=sum((metrics.key_rate_profile.get(key, Decimal(0)) * value for key, value in tenor_shocks_bps.items()), Decimal(0)),
        scenario_name=scenario_name or "key_rate_shift",
        breakdown=dict(metrics.key_rate_profile),
    )


def run_stress_scenarios(portfolio: Portfolio, *, curve, settlement_date, scenarios: list[object]) -> StressSummary:
    """Run a list of stress scenarios and return a summary."""

    results: dict[str, StressResult] = {}
    for scenario in scenarios:
        resolved = _run_stress_result(portfolio, curve=curve, settlement_date=settlement_date, scenario=scenario)
        if resolved is None:
            continue
        name, result = resolved
        results[name] = result
    return summarize_results(results)


def run_stress_scenario(portfolio: Portfolio, *, curve, settlement_date, scenario: object) -> StressResult:
    """Run a single stress scenario and return its result."""

    resolved = _run_stress_result(portfolio, curve=curve, settlement_date=settlement_date, scenario=scenario)
    if resolved is None:
        raise TypeError(f"Unsupported stress scenario: {scenario!r}")
    return resolved[1]


def stress_scenarios(portfolio: Portfolio, *, curve, settlement_date, scenarios: list[object]) -> StressSummary:
    """Compatibility alias for :func:`run_stress_scenarios`."""

    return run_stress_scenarios(portfolio, curve=curve, settlement_date=settlement_date, scenarios=scenarios)


__all__ = [
    "key_rate_shift_impact",
    "key_rate_shift_result",
    "parallel_shift_impact",
    "rate_shock_impact",
    "run_stress_scenario",
    "run_stress_scenarios",
    "spread_shock_impact",
    "spread_shock_result",
    "stress_scenarios",
]



from __future__ import annotations

import pytest

from fuggers_py.math.errors import (
    ConvergenceFailed,
    DivisionByZero,
    DimensionMismatch,
    ExtrapolationNotAllowed,
    InsufficientData,
    InvalidBracket,
    InvalidInput,
    MathError,
    MathOverflow,
    MathUnderflow,
    SingularMatrix,
)


def test_error_payloads_and_str() -> None:
    err = ConvergenceFailed(iterations=5, residual=1e-3)
    assert err.iterations == 5
    assert err.residual == 1e-3
    assert "Convergence failed" in str(err)

    br = InvalidBracket(a=0.0, b=1.0, fa=1.0, fb=2.0)
    assert br.a == 0.0
    assert "Invalid bracket" in str(br)

    dz = DivisionByZero(value=0.0)
    assert "Division by zero" in str(dz)

    sm = SingularMatrix()
    assert "Singular matrix" in str(sm)

    dm = DimensionMismatch(rows1=2, cols1=2, rows2=3, cols2=1)
    assert "Dimension mismatch" in str(dm)

    ex = ExtrapolationNotAllowed(x=2.0, min=0.0, max=1.0)
    assert "Extrapolation not allowed" in str(ex)

    ins = InsufficientData(required=2, actual=1)
    assert "Insufficient data" in str(ins)

    inv = InvalidInput(reason="bad")
    assert "Invalid input" in str(inv)

    of = MathOverflow(operation="exp")
    assert "overflow" in str(of).lower()

    uf = MathUnderflow(operation="exp")
    assert "underflow" in str(uf).lower()


def test_error_constructor_helpers() -> None:
    err = MathError.convergence_failed(10, 0.25)
    assert isinstance(err, ConvergenceFailed)
    assert err.iterations == 10

    err2 = MathError.invalid_input("nope")
    assert isinstance(err2, InvalidInput)
    assert "nope" in str(err2)

    err3 = MathError.insufficient_data(3, 1)
    assert isinstance(err3, InsufficientData)
    assert err3.required == 3


def test_invalid_input_is_exception() -> None:
    with pytest.raises(MathError):
        raise InvalidInput(reason="x")




from __future__ import annotations

import pytest

from fuggers_py.core import (
    BusinessDayConvention,
    CalendarData,
    CustomCalendarBuilder,
    DynamicCalendar,
    HolidayBitmap,
    HolidayCalendarBuilder,
    JapanCalendar,
    JointCalendar,
    SIFMACalendar,
    Target2Calendar,
    UKCalendar,
    USGovernmentCalendar,
    WeekendCalendar,
    WeekendType,
)
from fuggers_py.core.calendars import CalendarError, easter_sunday, last_weekday_of_month, nth_weekday_of_month, observed_date
from fuggers_py.core.types import Date


def test_weekend_type_is_weekend() -> None:
    assert WeekendType.SATURDAY_SUNDAY.is_weekend(5) is True
    assert WeekendType.SATURDAY_SUNDAY.is_weekend(6) is True
    assert WeekendType.SATURDAY_SUNDAY.is_weekend(0) is False
    assert WeekendType.NONE.is_weekend(6) is False
    assert str(WeekendType.SATURDAY_SUNDAY) == "Saturday Sunday"
    assert str(BusinessDayConvention.MODIFIED_FOLLOWING) == "Modified Following"


def test_business_day_adjustment_rules_on_weekend_calendar() -> None:
    cal = WeekendCalendar()
    sat = Date.parse("2025-03-01")
    assert cal.adjust(sat, BusinessDayConvention.FOLLOWING) == Date.parse("2025-03-03")
    assert cal.adjust(sat, BusinessDayConvention.PRECEDING) == Date.parse("2025-02-28")

    # Modified following crosses month for 2025-05-31 (Sat) -> 2025-06-02 (Mon) => go preceding.
    eom_sat = Date.parse("2025-05-31")
    assert cal.adjust(eom_sat, BusinessDayConvention.MODIFIED_FOLLOWING) == Date.parse("2025-05-30")

    # Modified preceding for 2025-03-01 (Sat) -> preceding 2025-02-28 (month changes) => go following.
    assert cal.adjust(sat, BusinessDayConvention.MODIFIED_PRECEDING) == Date.parse("2025-03-03")

    assert cal.adjust(sat, BusinessDayConvention.NEAREST) == Date.parse("2025-02-28")
    assert cal.adjust(eom_sat, BusinessDayConvention.END_OF_MONTH) == Date.parse("2025-05-30")


def test_weekend_calendar_business_days_between_semantics() -> None:
    cal = WeekendCalendar()
    start = Date.parse("2025-03-03")  # Monday
    end = Date.parse("2025-03-04")  # Tuesday
    assert cal.business_days_between(start, end) == 1


def test_calendar_helper_functions_cover_observed_dates_and_invalid_inputs() -> None:
    assert easter_sunday(2024) == Date.parse("2024-03-31")
    assert observed_date(Date.parse("2022-01-01")) == Date.parse("2021-12-31")
    assert observed_date(Date.parse("2023-01-01")) == Date.parse("2023-01-02")
    assert last_weekday_of_month(2024, 5, 0) == Date.parse("2024-05-27")
    assert nth_weekday_of_month(2024, 9, 0, 1) == Date.parse("2024-09-02")

    with pytest.raises(CalendarError, match="Invalid month"):
        last_weekday_of_month(2024, 13, 0)

    with pytest.raises(CalendarError, match="occurrence must be >= 1"):
        nth_weekday_of_month(2024, 1, 0, 0)

    with pytest.raises(CalendarError, match="does not exist"):
        nth_weekday_of_month(2024, 2, 0, 5)


def test_weekend_calendar_business_day_navigation_and_nearest_tie_break() -> None:
    cal = WeekendCalendar()
    start = Date.parse("2025-02-28")  # Friday

    assert cal.add_business_days(start, 1) == Date.parse("2025-03-03")
    assert cal.add_business_days(start, -1) == Date.parse("2025-02-27")
    assert cal.settlement_date(start, 1) == Date.parse("2025-03-03")
    assert cal.next_business_day(Date.parse("2025-03-01")) == Date.parse("2025-03-03")
    assert cal.previous_business_day(Date.parse("2025-03-02")) == Date.parse("2025-02-28")
    assert cal.business_days_between(Date.parse("2025-03-04"), Date.parse("2025-03-03")) == -1

    dyn = DynamicCalendar.from_calendar_data(CalendarData.new("Tie").with_holiday(Date.parse("2025-03-05")))
    assert dyn.adjust(Date.parse("2025-03-05"), BusinessDayConvention.NEAREST) == Date.parse("2025-03-06")


def test_sifma_known_holidays() -> None:
    sifma = SIFMACalendar.global_()
    assert sifma.is_business_day(Date.parse("2024-07-04")) is False  # Independence Day
    assert sifma.is_business_day(Date.parse("2024-03-29")) is False  # Good Friday (2024)
    assert sifma.is_business_day(Date.parse("2022-06-20")) is False  # Juneteenth observed (2022)
    assert sifma.is_business_day(Date.parse("2021-12-31")) is False  # New Year observed for 2022
    assert SIFMACalendar.global_() is sifma


def test_us_government_calendar_smoke() -> None:
    usg = USGovernmentCalendar.new()
    assert usg.is_business_day(Date.parse("2024-07-04")) is False


def test_target2_calendar_holidays() -> None:
    t2 = Target2Calendar.global_()
    assert t2.is_business_day(Date.parse("2024-05-01")) is False
    assert t2.is_business_day(Date.parse("2024-04-01")) is False  # Easter Monday 2024


def test_uk_calendar_holidays() -> None:
    uk = UKCalendar.global_()
    assert uk.is_business_day(Date.parse("2024-05-06")) is False  # Early May bank holiday 2024
    assert uk.is_business_day(Date.parse("2024-03-29")) is False  # Good Friday


def test_japan_calendar_holidays() -> None:
    jp = JapanCalendar.global_()
    assert jp.is_business_day(Date.parse("2024-01-02")) is False
    assert jp.is_business_day(Date.parse("2024-02-23")) is False
    assert jp.is_business_day(Date.parse("2024-01-08")) is False  # Coming of Age Day (2nd Monday)


def test_joint_calendar_semantics() -> None:
    joint = JointCalendar.new([SIFMACalendar.global_(), Target2Calendar.global_()])
    assert joint.is_business_day(Date.parse("2024-07-04")) is False  # closed in US
    assert joint.is_business_day(Date.parse("2024-05-01")) is False  # closed in TARGET2


def test_holiday_bitmap_and_builders() -> None:
    hb = HolidayBitmap.new("Test")
    d = Date.parse("2024-07-04")
    hb.add_holiday(d)
    assert hb.is_holiday(d) is True
    assert hb.is_business_day_date(d) is False
    hb.remove_holiday(d)
    assert hb.is_holiday(d) is False

    b = HolidayCalendarBuilder.new("Built").year_range(2024, 2024)
    b.add_fixed_holiday(7, 4, observe_weekend=True)
    built = b.build()
    assert built.is_business_day_date(Date.parse("2024-07-04")) is False


def test_calendar_data_and_dynamic_calendar_round_trip() -> None:
    data = CalendarData.new("Dyn").with_weekend(WeekendType.SATURDAY_SUNDAY).with_holiday(Date.parse("2024-07-04"))
    cal = DynamicCalendar.from_calendar_data(data)
    assert cal.is_business_day(Date.parse("2024-07-04")) is False

    text = cal.to_json()
    cal2 = DynamicCalendar.from_json(text)
    assert cal2.is_business_day(Date.parse("2024-07-04")) is False
    assert cal2.holiday_count() == 1


def test_custom_calendar_builder() -> None:
    cal = (
        CustomCalendarBuilder.new("Custom")
        .year_range(2024, 2024)
        .add_fixed_holiday_observed(7, 4)
        .add_good_friday()
        .build()
    )
    assert cal.is_business_day(Date.parse("2024-07-04")) is False
    assert cal.is_business_day(Date.parse("2024-03-29")) is False



"""Credit instruments.

This package exposes CDS instruments and their protection-side sign
conventions.
"""

from __future__ import annotations

from .cds import Cds, CdsPremiumPeriod, CreditDefaultSwap, ProtectionSide

__all__ = [
    "Cds",
    "CdsPremiumPeriod",
    "CreditDefaultSwap",
    "ProtectionSide",
]



from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal

import pytest

from fuggers_py.analytics.spreads import (
    GQDSecuredUnsecuredBasisModel,
    SecuredUnsecuredBasisModel,
    adjusted_term_rate,
    compounding_convexity_breakdown,
    reference_rate_decomposition,
    secured_unsecured_overnight_basis,
)


def test_reference_rate_decomposition_uses_explicit_ladder_sign_convention() -> None:
    breakdown = reference_rate_decomposition(
        repo_rate=Decimal("0.0310"),
        general_collateral_rate=Decimal("0.0320"),
        unsecured_overnight_rate=Decimal("0.0340"),
        term_rate=Decimal("0.0370"),
        convexity_adjustment=Decimal("0.0002"),
    )

    assert breakdown.repo_vs_gc == Decimal("0.0010")
    assert breakdown.gc_vs_unsecured_overnight == Decimal("0.0020")
    assert breakdown.unsecured_overnight_vs_term == Decimal("0.0032")
    assert breakdown.total_funding_basis == Decimal("0.0062")
    assert breakdown.adjusted_term_rate == Decimal("0.0372")


def test_compounding_convexity_breakdown_and_adjusted_term_rate_are_consistent() -> None:
    breakdown = compounding_convexity_breakdown(
        simple_rate=Decimal("0.0500"),
        year_fraction=Decimal("0.50"),
        convexity_adjustment=Decimal("0.0001"),
    )

    assert breakdown.compounded_equivalent_rate > breakdown.simple_rate
    assert breakdown.compounding_adjustment == breakdown.compounded_equivalent_rate - breakdown.simple_rate
    assert breakdown.adjusted_term_rate == adjusted_term_rate(
        simple_rate=Decimal("0.0500"),
        year_fraction=Decimal("0.50"),
        convexity_adjustment=Decimal("0.0001"),
    )


def test_secured_unsecured_overnight_basis_defaults_to_gqd_model() -> None:
    basis = secured_unsecured_overnight_basis(
        loss_given_default=Decimal("0.60"),
        default_probability=Decimal("0.02"),
        discount_factor=Decimal("0.99"),
    )

    assert basis == GQDSecuredUnsecuredBasisModel().basis(
        loss_given_default=Decimal("0.60"),
        default_probability=Decimal("0.02"),
        discount_factor=Decimal("0.99"),
    )
    assert basis == Decimal("0.011880")


def test_secured_unsecured_overnight_basis_accepts_custom_model() -> None:
    @dataclass(frozen=True, slots=True)
    class FlatBasisModel(SecuredUnsecuredBasisModel):
        value: Decimal

        def basis(
            self,
            *,
            loss_given_default: object,
            default_probability: object,
            discount_factor: object,
        ) -> Decimal:
            return self.value

    basis = secured_unsecured_overnight_basis(
        loss_given_default=Decimal("0.60"),
        default_probability=Decimal("0.02"),
        discount_factor=Decimal("0.99"),
        model=FlatBasisModel(Decimal("0.0015")),
    )

    assert basis == Decimal("0.0015")


def test_secured_unsecured_basis_validates_gqd_inputs() -> None:
    with pytest.raises(ValueError, match="loss_given_default"):
        secured_unsecured_overnight_basis(
            loss_given_default=Decimal("1.10"),
            default_probability=Decimal("0.02"),
            discount_factor=Decimal("0.99"),
        )



from __future__ import annotations

from dataclasses import replace
from decimal import Decimal

from fuggers_py.bonds.indices import BondIndex, IndexConventions, IndexFixingStore, OvernightCompounding
from fuggers_py.bonds.instruments import CallableBondBuilder, FixedBond, FloatingRateNoteBuilder
from fuggers_py.bonds.types import RateIndex, YieldCalculationRules
from fuggers_py.core import Currency, Date, Frequency
from fuggers_py.engine import CurveBuilder, PortfolioPosition, PricingRouter
from fuggers_py.data import (
    AnalyticsCurves,
    BondReferenceData,
    BondType,
    CallScheduleEntry,
    CurveId,
    CurvePoint,
    EtfHolding,
    InstrumentId,
    InMemoryFixingSource,
    IndexFixing,
    IssuerType,
    MarketDataSnapshot,
    PricingSpec,
    RawQuote,
    QuoteSide,
    FloatingRateTerms,
)


SETTLEMENT = Date.from_ymd(2026, 1, 15)
FIXED_ID = InstrumentId("SCENARIO-A-FIXED")
CALLABLE_ID = InstrumentId("SCENARIO-B-CALLABLE")
FRN_ID = InstrumentId("SCENARIO-C-FRN")
ETF_ID = "SCENARIO-D-ETF"
PORTFOLIO_ID = "SCENARIO-E-PORTFOLIO"

DISCOUNT_ID = CurveId("usd.discount")
GOVERNMENT_ID = CurveId("usd.government")
BENCHMARK_ID = CurveId("usd.benchmark")
FORWARD_ID = CurveId("usd.forward")


def scenario_a_curve_points() -> list[CurvePoint]:
    return [
        CurvePoint(Decimal("0.25"), Decimal("0.0450")),
        CurvePoint(Decimal("0.50"), Decimal("0.0445")),
        CurvePoint(Decimal("1.00"), Decimal("0.0435")),
        CurvePoint(Decimal("2.00"), Decimal("0.0415")),
        CurvePoint(Decimal("5.00"), Decimal("0.0395")),
        CurvePoint(Decimal("10.00"), Decimal("0.0400")),
    ]


def scenario_a_government_points() -> list[CurvePoint]:
    return [
        CurvePoint(Decimal("0.25"), Decimal("0.0420")),
        CurvePoint(Decimal("0.50"), Decimal("0.0410")),
        CurvePoint(Decimal("1.00"), Decimal("0.0400")),
        CurvePoint(Decimal("2.00"), Decimal("0.0385")),
        CurvePoint(Decimal("5.00"), Decimal("0.0370")),
        CurvePoint(Decimal("10.00"), Decimal("0.0380")),
    ]


def scenario_a_benchmark_points() -> list[CurvePoint]:
    return [
        CurvePoint(Decimal("0.25"), Decimal("0.0435")),
        CurvePoint(Decimal("0.50"), Decimal("0.0428")),
        CurvePoint(Decimal("1.00"), Decimal("0.0418")),
        CurvePoint(Decimal("2.00"), Decimal("0.0402")),
        CurvePoint(Decimal("5.00"), Decimal("0.0388")),
        CurvePoint(Decimal("10.00"), Decimal("0.0392")),
    ]


def scenario_c_curve_points() -> list[CurvePoint]:
    return [
        CurvePoint(Decimal("0.25"), Decimal("0.0430")),
        CurvePoint(Decimal("0.50"), Decimal("0.0420")),
        CurvePoint(Decimal("1.00"), Decimal("0.0405")),
        CurvePoint(Decimal("2.00"), Decimal("0.0390")),
        CurvePoint(Decimal("3.00"), Decimal("0.0385")),
    ]


def curve_builder_with_scenarios() -> CurveBuilder:
    builder = CurveBuilder()
    builder.add_zero_curve(DISCOUNT_ID, scenario_a_curve_points(), SETTLEMENT)
    builder.add_zero_curve(GOVERNMENT_ID, scenario_a_government_points(), SETTLEMENT)
    builder.add_zero_curve(BENCHMARK_ID, scenario_a_benchmark_points(), SETTLEMENT)
    builder.add_zero_curve(CurveId("frn.discount"), scenario_c_curve_points(), SETTLEMENT)
    builder.add_forward_curve(FORWARD_ID, scenario_c_curve_points(), SETTLEMENT)
    return builder


def scenario_a_instrument() -> FixedBond:
    rules = replace(YieldCalculationRules.us_corporate(), frequency=Frequency.SEMI_ANNUAL)
    return FixedBond.new(
        issue_date=Date.from_ymd(2024, 1, 15),
        maturity_date=Date.from_ymd(2031, 1, 15),
        coupon_rate=Decimal("0.0450"),
        frequency=Frequency.SEMI_ANNUAL,
        currency=Currency.USD,
        notional=Decimal("100"),
        rules=rules,
    )


def scenario_b_instrument():
    rules = replace(YieldCalculationRules.us_corporate(), frequency=Frequency.SEMI_ANNUAL)
    base = FixedBond.new(
        issue_date=Date.from_ymd(2023, 1, 15),
        maturity_date=Date.from_ymd(2033, 1, 15),
        coupon_rate=Decimal("0.0500"),
        frequency=Frequency.SEMI_ANNUAL,
        currency=Currency.USD,
        notional=Decimal("100"),
        rules=rules,
    )
    builder = CallableBondBuilder.new().with_base_bond(base)
    for year in range(2028, 2033):
        builder = builder.add_call(call_date=Date.from_ymd(year, 1, 15), call_price=Decimal("100.00"))
    return builder.build()


def scenario_c_fixing_source() -> InMemoryFixingSource:
    fixings = [
        IndexFixing("SOFR", Date.from_ymd(2025, 12, 29), Decimal("0.0408")),
        IndexFixing("SOFR", Date.from_ymd(2025, 12, 30), Decimal("0.0409")),
        IndexFixing("SOFR", Date.from_ymd(2025, 12, 31), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 2), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 5), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 6), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 7), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 8), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 9), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 12), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 13), Decimal("0.0410")),
        IndexFixing("SOFR", Date.from_ymd(2026, 1, 14), Decimal("0.0410")),
    ]
    return InMemoryFixingSource(fixings)


def scenario_c_instrument():
    store = scenario_c_fixing_source().to_fixing_store()
    index = BondIndex(
        name="SOFR",
        rate_index=RateIndex.SOFR,
        currency=Currency.USD,
        fixing_store=store,
        conventions=IndexConventions(overnight_compounding=OvernightCompounding.COMPOUNDED),
    )
    rules = replace(YieldCalculationRules.us_treasury(), frequency=Frequency.QUARTERLY)
    return (
        FloatingRateNoteBuilder.new()
        .with_issue_date(Date.from_ymd(2025, 1, 15))
        .with_maturity_date(Date.from_ymd(2028, 1, 15))
        .with_index(RateIndex.SOFR)
        .with_index_definition(index)
        .with_quoted_spread(Decimal("0.0125"))
        .with_frequency(Frequency.QUARTERLY)
        .with_currency(Currency.USD)
        .with_notional(Decimal("100"))
        .with_rules(rules)
        .with_current_reference_rate(Decimal("0.0410"))
        .build()
    )


def fixed_curves() -> AnalyticsCurves:
    builder = curve_builder_with_scenarios()
    return AnalyticsCurves(
        discount_curve=builder.get(DISCOUNT_ID),
        government_curve=builder.get(GOVERNMENT_ID),
        benchmark_curve=builder.get(BENCHMARK_ID),
    )


def frn_curves() -> AnalyticsCurves:
    builder = curve_builder_with_scenarios()
    return AnalyticsCurves(
        discount_curve=builder.get("frn.discount"),
        forward_curve=builder.get(FORWARD_ID),
    )


def scenario_market_data() -> MarketDataSnapshot:
    return MarketDataSnapshot(
        as_of=SETTLEMENT,
        quotes=(
            RawQuote(FIXED_ID, Decimal("101.25"), QuoteSide.MID, as_of=SETTLEMENT, currency=Currency.USD),
            RawQuote(CALLABLE_ID, Decimal("102.50"), QuoteSide.MID, as_of=SETTLEMENT, currency=Currency.USD),
            RawQuote(FRN_ID, Decimal("100.15"), QuoteSide.MID, as_of=SETTLEMENT, currency=Currency.USD),
        ),
        fixings=tuple(scenario_c_fixing_source().fixings.values()),
    )


def scenario_reference_data() -> dict[InstrumentId, BondReferenceData]:
    return {
        FIXED_ID: BondReferenceData(
            instrument_id=FIXED_ID,
            bond_type=BondType.FIXED_RATE,
            issuer_type=IssuerType.CORPORATE,
            issue_date=Date.from_ymd(2024, 1, 15),
            maturity_date=Date.from_ymd(2031, 1, 15),
            currency=Currency.USD,
            coupon_rate=Decimal("0.0450"),
            frequency=Frequency.SEMI_ANNUAL,
            sector="CORPORATE",
            rating="A",
        ),
        CALLABLE_ID: BondReferenceData(
            instrument_id=CALLABLE_ID,
            bond_type=BondType.CALLABLE,
            issuer_type=IssuerType.CORPORATE,
            issue_date=Date.from_ymd(2023, 1, 15),
            maturity_date=Date.from_ymd(2033, 1, 15),
            currency=Currency.USD,
            coupon_rate=Decimal("0.0500"),
            frequency=Frequency.SEMI_ANNUAL,
            call_schedule=tuple(
                CallScheduleEntry(Date.from_ymd(year, 1, 15), Decimal("100.00")) for year in range(2028, 2033)
            ),
            sector="FINANCIALS",
            rating="BBB",
        ),
        FRN_ID: BondReferenceData(
            instrument_id=FRN_ID,
            bond_type=BondType.FLOATING_RATE,
            issuer_type=IssuerType.CORPORATE,
            issue_date=Date.from_ymd(2025, 1, 15),
            maturity_date=Date.from_ymd(2028, 1, 15),
            currency=Currency.USD,
            floating_rate_terms=FloatingRateTerms(
                index_name="SOFR",
                spread=Decimal("0.0125"),
                reset_frequency=Frequency.QUARTERLY,
                current_reference_rate=Decimal("0.0410"),
            ),
            sector="CORPORATE",
            rating="A",
        ),
    }


def etf_holdings() -> list[EtfHolding]:
    return [
        EtfHolding(FIXED_ID, quantity=Decimal("100")),
        EtfHolding(CALLABLE_ID, quantity=Decimal("80")),
        EtfHolding(FRN_ID, quantity=Decimal("120")),
    ]


def portfolio_positions() -> list[PortfolioPosition]:
    return [
        PortfolioPosition(FIXED_ID, Decimal("100")),
        PortfolioPosition(CALLABLE_ID, Decimal("80")),
        PortfolioPosition(FRN_ID, Decimal("120")),
    ]


def router() -> PricingRouter:
    return PricingRouter()


def pricing_specs() -> tuple[PricingSpec, PricingSpec]:
    fixed = PricingSpec(compute_key_rates=True, include_asset_swap=True)
    floating = PricingSpec(market_price_is_dirty=True)
    return fixed, floating



from __future__ import annotations

from decimal import Decimal

import pytest

from fuggers_py.analytics.spreads import OASCalculator
from fuggers_py.bonds.instruments import CallType, CallableBondBuilder, FixedBond
from fuggers_py.bonds.options import HullWhiteModel
from fuggers_py.core import Date, Frequency
from fuggers_py.curves import DiscountCurveBuilder


def _callable_setup():
    settlement = Date.from_ymd(2024, 1, 1)
    base = FixedBond.new(
        issue_date=settlement,
        maturity_date=settlement.add_years(5),
        coupon_rate=Decimal("0.05"),
        frequency=Frequency.SEMI_ANNUAL,
    )
    callable_bond = (
        CallableBondBuilder.new()
        .with_base_bond(base)
        .add_call(call_date=settlement.add_years(2), call_price=Decimal("101"), call_type=CallType.EUROPEAN)
        .add_call(call_date=settlement.add_years(3), call_price=Decimal("100.5"), call_type=CallType.EUROPEAN)
        .build()
    )
    curve = (
        DiscountCurveBuilder(reference_date=settlement)
        .add_zero_rate(1.0, Decimal("0.04"))
        .add_zero_rate(10.0, Decimal("0.04"))
        .build()
    )
    model = HullWhiteModel(mean_reversion=Decimal("0.03"), volatility=Decimal("0.01"), term_structure=curve)
    return settlement, callable_bond, OASCalculator(model=model)


def test_oas_roundtrip() -> None:
    settlement, callable_bond, calculator = _callable_setup()
    oas = Decimal("0.005")
    price = calculator.price_with_oas(callable_bond, oas, settlement)
    recovered = calculator.calculate(callable_bond, price, settlement)
    assert float(recovered) == pytest.approx(float(oas), abs=1e-8)


def test_effective_duration_convexity_positive() -> None:
    settlement, callable_bond, calculator = _callable_setup()
    duration = calculator.effective_duration(callable_bond, Decimal("0.005"), settlement)
    convexity = calculator.effective_convexity(callable_bond, Decimal("0.005"), settlement)
    assert duration > 0
    assert convexity > 0
    assert float(duration) < 10


def test_option_value_sign_behavior() -> None:
    settlement, callable_bond, calculator = _callable_setup()
    option_value = calculator.option_value(callable_bond, Decimal("0.005"), settlement)
    callable_price = calculator.price_with_oas(callable_bond, Decimal("0.005"), settlement)
    assert option_value > 0
    assert callable_price > 0



"""Sinking-fund bond instruments.

Sinking-fund factors represent the remaining principal fraction outstanding
after each scheduled amortization date. Coupon rates are raw decimal rates.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from decimal import Decimal

from fuggers_py.core.types import Currency, Date, Frequency, Price
from fuggers_py.math import SolverConfig, brent
from fuggers_py.math.errors import ConvergenceFailed, InvalidBracket

from ..cashflows import AccruedInterestCalculator, AccruedInterestInputs, CashFlowGenerator, Schedule, ScheduleConfig
from ..errors import InvalidBondSpec
from ..pricing import StandardYieldEngine
from ..traits import Bond, BondAnalytics, BondCashFlow, CashFlowType
from ..types import AccruedConvention, BondIdentifiers, StubPeriodRules, YieldCalculationRules


def _to_decimal(value: object, *, field_name: str) -> Decimal:
    if isinstance(value, Decimal):
        return value
    if isinstance(value, (int, str)):
        return Decimal(value)
    if isinstance(value, float):
        return Decimal(str(value))
    return Decimal(str(value))


@dataclass(frozen=True, slots=True)
class SinkingFundEntry:
    """One scheduled sinking-fund factor reset."""

    payment_date: Date
    factor: Decimal

    def __post_init__(self) -> None:
        object.__setattr__(self, "factor", _to_decimal(self.factor, field_name="factor"))
        if self.factor < 0 or self.factor > 1:
            raise InvalidBondSpec(reason="SinkingFundEntry.factor must be between 0 and 1.")


@dataclass(frozen=True, slots=True)
class SinkingFundSchedule:
    """Ordered sinking-fund factors with non-increasing outstanding principal."""

    entries: tuple[SinkingFundEntry, ...]

    def __post_init__(self) -> None:
        ordered = tuple(sorted(self.entries, key=lambda entry: entry.payment_date))
        if not ordered:
            raise InvalidBondSpec(reason="SinkingFundSchedule requires at least one entry.")
        previous_factor = Decimal(1)
        for entry in ordered:
            if entry.factor > previous_factor:
                raise InvalidBondSpec(reason="Sinking fund factors must be non-increasing.")
            previous_factor = entry.factor
        object.__setattr__(self, "entries", ordered)

    @classmethod
    def new(cls, entries: list[SinkingFundEntry]) -> "SinkingFundSchedule":
        """Create a sinking-fund schedule from the supplied entries."""
        return cls(entries=tuple(entries))

    def factor_on(self, date: Date) -> Decimal:
        """Return the remaining principal factor in force on ``date``."""
        factor = Decimal(1)
        for entry in self.entries:
            if date >= entry.payment_date:
                factor = entry.factor
            else:
                break
        return factor

    def factor_for_payment(self, payment_date: Date, *, default: Decimal = Decimal(0)) -> Decimal:
        """Return the factor applied on the payment date, if scheduled."""
        for entry in self.entries:
            if entry.payment_date == payment_date:
                return entry.factor
        return default

    def to_amortization(self, *, notional: Decimal) -> list[tuple[Date, Decimal]]:
        """Convert remaining-principal factors into principal amortization cash flows."""
        amortization: list[tuple[Date, Decimal]] = []
        prior_factor = Decimal(1)
        for entry in self.entries:
            principal = (prior_factor - entry.factor) * notional
            amortization.append((entry.payment_date, principal))
            prior_factor = entry.factor
        return amortization


@dataclass(frozen=True, slots=True)
class SinkingFundBond(BondAnalytics, Bond):
    """Fixed-rate bond with scheduled principal amortization."""

    _issue_date: Date
    _maturity_date: Date
    _coupon_rate: Decimal
    _frequency: Frequency
    _sinking_schedule: SinkingFundSchedule
    _currency: Currency = Currency.USD
    _notional: Decimal = Decimal(100)
    _identifiers: BondIdentifiers = field(default_factory=BondIdentifiers)
    _rules: YieldCalculationRules = field(default_factory=YieldCalculationRules.us_treasury)
    _stub_rules: StubPeriodRules | None = None

    _schedule: Schedule = field(init=False, repr=False)
    _cashflows: list[BondCashFlow] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        if self._maturity_date <= self._issue_date:
            raise InvalidBondSpec(reason="maturity_date must be after issue_date.")
        if self._frequency.is_zero():
            raise InvalidBondSpec(reason="SinkingFundBond requires a non-zero coupon frequency.")
        if self._notional <= 0:
            raise InvalidBondSpec(reason="notional must be positive.")
        rules = self.rules()
        if rules.frequency != self._frequency:
            raise InvalidBondSpec(
                reason="Bond frequency must match rules.frequency (set via YieldCalculationRules)."
            )
        stub_rules = self._stub_rules or rules.stub_rules
        schedule_config = ScheduleConfig(
            start_date=self._issue_date,
            end_date=self._maturity_date,
            frequency=self._frequency,
            calendar=rules.calendar,
            business_day_convention=rules.business_day_convention,
            end_of_month=rules.end_of_month,
            stub_rules=stub_rules,
        )
        schedule = Schedule.generate(schedule_config)
        cashflows = self._build_cashflows(schedule)
        object.__setattr__(self, "_schedule", schedule)
        object.__setattr__(self, "_cashflows", cashflows)

    @classmethod
    def new(
        cls,
        *,
        issue_date: Date,
        maturity_date: Date,
        coupon_rate: Decimal,
        frequency: Frequency,
        sinking_schedule: SinkingFundSchedule,
        currency: Currency = Currency.USD,
        notional: Decimal = Decimal(100),
        identifiers: BondIdentifiers | None = None,
        rules: YieldCalculationRules | None = None,
        stub_rules: StubPeriodRules | None = None,
    ) -> "SinkingFundBond":
        """Construct a sinking-fund bond from its contractual inputs."""
        return cls(
            _issue_date=issue_date,
            _maturity_date=maturity_date,
            _coupon_rate=coupon_rate,
            _frequency=frequency,
            _sinking_schedule=sinking_schedule,
            _currency=currency,
            _notional=notional,
            _identifiers=identifiers or BondIdentifiers(),
            _rules=rules or YieldCalculationRules.us_treasury(),
            _stub_rules=stub_rules,
        )

    def identifiers(self) -> BondIdentifiers:
        return self._identifiers

    def currency(self) -> Currency:
        return self._currency

    def notional(self) -> Decimal:
        return self._notional

    def issue_date(self) -> Date:
        return self._issue_date

    def maturity_date(self) -> Date:
        return self._maturity_date

    def frequency(self) -> Frequency:
        return self._frequency

    def coupon_rate(self) -> Decimal:
        return self._coupon_rate

    def rules(self) -> YieldCalculationRules:
        return self._rules

    def schedule(self) -> Schedule:
        return self._schedule

    def sinking_schedule(self) -> SinkingFundSchedule:
        return self._sinking_schedule

    def cash_flows(self, from_date: Date | None = None) -> list[BondCashFlow]:
        return CashFlowGenerator.future_cashflows(self._cashflows, from_date)

    def accrued_interest(self, settlement_date: Date) -> Decimal:
        """Return accrued interest adjusted for the outstanding sinking factor."""
        if settlement_date >= self._maturity_date:
            return Decimal(0)

        rules = self.rules()
        day_count = rules.accrual_day_count_obj()
        unadjusted = self._schedule.unadjusted_dates
        adjusted = self._schedule.dates

        for i in range(1, len(unadjusted)):
            accrual_start = unadjusted[i - 1]
            accrual_end = unadjusted[i]
            if not (accrual_start < settlement_date < accrual_end):
                continue
            factor = self.factor_on(accrual_start)
            coupon_amount = self._notional * factor * self._coupon_rate * day_count.year_fraction(accrual_start, accrual_end)
            coupon_date = adjusted[i]
            inputs = AccruedInterestInputs(
                settlement_date=settlement_date,
                accrual_start=accrual_start,
                accrual_end=accrual_end,
                coupon_amount=coupon_amount,
                coupon_date=coupon_date,
                full_coupon_amount=coupon_amount,
                period_start=accrual_start,
                period_end=accrual_end,
            )
            if rules.accrued_convention is AccruedConvention.EX_DIVIDEND:
                return AccruedInterestCalculator.ex_dividend(inputs, rules=rules)
            return AccruedInterestCalculator.standard(inputs, rules=rules)
        return Decimal(0)

    def factor_on(self, date: Date) -> Decimal:
        """Return the remaining principal factor on ``date``."""
        return self._sinking_schedule.factor_on(date)

    def amortization_schedule(self) -> list[tuple[Date, Decimal]]:
        """Return scheduled principal reductions in currency or face units."""
        return self._sinking_schedule.to_amortization(notional=self._notional)

    def average_life(self, *, as_of_date: Date | None = None) -> Decimal:
        """Return the weighted-average time to remaining principal payments, in years."""
        anchor = as_of_date or self.issue_date()
        amortization = [(date, amount) for date, amount in self.amortization_schedule() if date > anchor and amount > 0]
        remaining_principal = sum(amount for _, amount in amortization)
        if remaining_principal == 0:
            return Decimal(0)
        weighted_years = Decimal(0)
        for payment_date, principal_amount in amortization:
            years = Decimal(str(float(anchor.days_between(payment_date)) / 365.0))
            weighted_years += years * principal_amount
        return weighted_years / remaining_principal

    def yield_to_average_life(self, clean_price: object, settlement_date: Date) -> Decimal:
        """Return the raw decimal yield implied by the average-life cash flows."""
        price = clean_price if isinstance(clean_price, Price) else Price.new(clean_price, self.currency())
        dirty_price = price.as_percentage() + self.accrued_interest(settlement_date)
        cashflows = [cf for cf in self.cash_flows(settlement_date)]
        if not cashflows:
            raise InvalidBondSpec(reason="No future cash flows remain for yield calculation.")

        day_count = self.rules().yield_day_count_obj()
        compounding = self.rules().compounding

        def objective(y: float) -> float:
            pv = 0.0
            for cf in cashflows:
                years = float(day_count.year_fraction(settlement_date, cf.date))
                pv += float(cf.factored_amount()) * compounding.discount_factor(y, years)
            return pv - float(dirty_price)

        config = SolverConfig(tolerance=1e-10, max_iterations=200)
        for lower, upper in [(-0.5, 0.5), (-0.9, 1.0), (-0.99, 2.0)]:
            try:
                result = brent(objective, lower, upper, config=config)
                return Decimal(str(result.root))
            except (ConvergenceFailed, InvalidBracket):
                continue
        raise InvalidBondSpec(reason="yield_to_average_life failed to converge.")

    def _build_cashflows(self, schedule: Schedule) -> list[BondCashFlow]:
        """Build coupon and principal cash flows using the sinking-fund factors."""
        rules = self.rules()
        day_count = rules.accrual_day_count_obj()
        flows: list[BondCashFlow] = []
        outstanding_factor = Decimal(1)
        for index in range(1, len(schedule.unadjusted_dates)):
            accrual_start = schedule.unadjusted_dates[index - 1]
            accrual_end = schedule.unadjusted_dates[index]
            pay_date = schedule.dates[index]
            accrual_factor = day_count.year_fraction(accrual_start, accrual_end)
            coupon_amount = self._notional * outstanding_factor * self._coupon_rate * accrual_factor
            ending_factor = self._sinking_schedule.factor_for_payment(
                accrual_end,
                default=Decimal(0) if accrual_end == self._maturity_date else outstanding_factor,
            )
            principal_amount = self._notional * (outstanding_factor - ending_factor)
            amount = coupon_amount + principal_amount
            flow_type = (
                CashFlowType.COUPON_AND_PRINCIPAL if principal_amount > 0 else CashFlowType.COUPON
            )
            flows.append(
                BondCashFlow(
                    date=pay_date,
                    amount=amount,
                    flow_type=flow_type,
                    accrual_start=accrual_start,
                    accrual_end=accrual_end,
                )
            )
            outstanding_factor = ending_factor
        return flows


@dataclass(slots=True)
class SinkingFundBondBuilder:
    """Fluent builder for :class:`SinkingFundBond`."""

    issue_date: Date | None = None
    maturity_date: Date | None = None
    coupon_rate: Decimal | None = None
    frequency: Frequency | None = None
    currency: Currency = Currency.USD
    notional: Decimal = Decimal(100)
    identifiers: BondIdentifiers = field(default_factory=BondIdentifiers)
    rules: YieldCalculationRules = field(default_factory=YieldCalculationRules.us_treasury)
    stub_rules: StubPeriodRules | None = None
    sinking_schedule: SinkingFundSchedule | None = None
    _entries: list[SinkingFundEntry] = field(default_factory=list, repr=False)

    @classmethod
    def new(cls) -> "SinkingFundBondBuilder":
        return cls()

    def with_issue_date(self, date: Date) -> "SinkingFundBondBuilder":
        self.issue_date = date
        return self

    def with_maturity_date(self, date: Date) -> "SinkingFundBondBuilder":
        self.maturity_date = date
        return self

    def with_coupon_rate(self, rate: object) -> "SinkingFundBondBuilder":
        self.coupon_rate = _to_decimal(rate, field_name="coupon_rate")
        return self

    def with_frequency(self, frequency: Frequency) -> "SinkingFundBondBuilder":
        self.frequency = frequency
        return self

    def with_currency(self, currency: Currency) -> "SinkingFundBondBuilder":
        self.currency = currency
        return self

    def with_notional(self, notional: object) -> "SinkingFundBondBuilder":
        self.notional = _to_decimal(notional, field_name="notional")
        return self

    def with_identifiers(self, identifiers: BondIdentifiers) -> "SinkingFundBondBuilder":
        self.identifiers = identifiers
        return self

    def with_rules(self, rules: YieldCalculationRules) -> "SinkingFundBondBuilder":
        self.rules = rules
        if self.frequency is None:
            self.frequency = rules.frequency
        return self

    def with_stub_rules(self, stub_rules: StubPeriodRules) -> "SinkingFundBondBuilder":
        self.stub_rules = stub_rules
        return self

    def with_sinking_schedule(self, schedule: SinkingFundSchedule) -> "SinkingFundBondBuilder":
        self.sinking_schedule = schedule
        self._entries = list(schedule.entries)
        return self

    def add_sinking_entry(self, payment_date: Date, factor: object) -> "SinkingFundBondBuilder":
        self._entries.append(SinkingFundEntry(payment_date=payment_date, factor=_to_decimal(factor, field_name="factor")))
        self.sinking_schedule = None
        return self

    def build_schedule(self) -> SinkingFundSchedule:
        if self.sinking_schedule is not None:
            return self.sinking_schedule
        if not self._entries:
            raise InvalidBondSpec(reason="sinking_schedule is required.")
        self.sinking_schedule = SinkingFundSchedule.new(self._entries)
        return self.sinking_schedule

    def build(self) -> SinkingFundBond:
        if self.issue_date is None:
            raise InvalidBondSpec(reason="issue_date is required.")
        if self.maturity_date is None:
            raise InvalidBondSpec(reason="maturity_date is required.")
        if self.coupon_rate is None:
            raise InvalidBondSpec(reason="coupon_rate is required.")
        frequency = self.frequency or self.rules.frequency
        schedule = self.build_schedule()
        return SinkingFundBond.new(
            issue_date=self.issue_date,
            maturity_date=self.maturity_date,
            coupon_rate=self.coupon_rate,
            frequency=frequency,
            sinking_schedule=schedule,
            currency=self.currency,
            notional=self.notional,
            identifiers=self.identifiers,
            rules=self.rules,
            stub_rules=self.stub_rules,
        )


__all__ = ["SinkingFundBond", "SinkingFundBondBuilder", "SinkingFundEntry", "SinkingFundSchedule"]



"""Multi-factor stochastic delivery-option model.

Each scenario supplies factor moves in dimensionless shock units that are
scaled by factor volatilities expressed in basis points and then mapped through
instrument-specific factor loadings.
"""

from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal

from fuggers_py.bonds.pricing import BondPricer
from fuggers_py.data.ids import InstrumentId

from ..conversion_factor import conversion_factor
from ..deliverable_basket import DeliverableBasket
from ..delivery_option import DeliveryOptionResult, DeliveryOptionScenario, _scenario_equivalent_prices, _to_decimal
from ..government_bond_future import GovernmentBondFuture


@dataclass(frozen=True, slots=True)
class MultiFactorScenario:
    """Single multi-factor delivery-option scenario."""

    factor_moves: tuple[Decimal, ...]
    probability: Decimal = Decimal(1)
    label: str | None = None

    def __post_init__(self) -> None:
        object.__setattr__(self, "factor_moves", tuple(_to_decimal(item) for item in self.factor_moves))
        object.__setattr__(self, "probability", _to_decimal(self.probability))
        if self.probability <= Decimal(0):
            raise ValueError("MultiFactorScenario probability must be positive.")
        if self.label is not None:
            object.__setattr__(self, "label", self.label.strip())


@dataclass(frozen=True, slots=True)
class _ScenarioCandidate:
    """Internal candidate ranked by futures-equivalent price."""

    instrument_id: InstrumentId
    futures_equivalent_price: Decimal


def _normalized_scenarios(scenarios: tuple[MultiFactorScenario, ...]) -> tuple[MultiFactorScenario, ...]:
    """Return scenarios with probabilities normalized to one."""
    total = sum((scenario.probability for scenario in scenarios), start=Decimal(0))
    return tuple(
        MultiFactorScenario(
            factor_moves=scenario.factor_moves,
            probability=scenario.probability / total,
            label=scenario.label,
        )
        for scenario in scenarios
    )


def _scenario_candidates_with_instrument_shifts(
    contract: GovernmentBondFuture,
    basket: DeliverableBasket,
    *,
    yield_shifts_bps: dict[InstrumentId, Decimal],
    prefer_published_conversion_factor: bool,
    pricer: BondPricer,
) -> tuple[_ScenarioCandidate, ...]:
    """Return ranked candidates after applying instrument-specific shifts."""
    delivery_date = contract.resolved_delivery_date()
    candidates: list[_ScenarioCandidate] = []
    for deliverable in basket.deliverables:
        shift = yield_shifts_bps[deliverable.instrument_id]
        clean_price = deliverable.price_with_yield_shift(
            delivery_date,
            base_settlement_date=basket.as_of,
            yield_shift_bps=shift,
            pricer=pricer,
        )
        cf = conversion_factor(
            contract,
            deliverable,
            prefer_published_override=prefer_published_conversion_factor,
            pricer=pricer,
        ).conversion_factor
        candidates.append(
            _ScenarioCandidate(
                instrument_id=deliverable.instrument_id,
                futures_equivalent_price=clean_price / cf,
            )
        )
    return tuple(sorted(candidates, key=lambda item: (item.futures_equivalent_price, item.instrument_id.as_str())))


@dataclass(frozen=True, slots=True)
class MultiFactorDeliveryOptionModel:
    """Delivery-option model driven by multiple factor shocks."""

    factor_volatilities_bps: tuple[Decimal, ...]
    deliverable_factor_loadings: dict[InstrumentId, tuple[Decimal, ...]]
    scenarios: tuple[MultiFactorScenario, ...]
    mean_shift_bps: Decimal = Decimal(0)

    def __post_init__(self) -> None:
        factor_volatilities = tuple(_to_decimal(item) for item in self.factor_volatilities_bps)
        if not factor_volatilities:
            raise ValueError("MultiFactorDeliveryOptionModel requires at least one factor volatility.")
        object.__setattr__(self, "factor_volatilities_bps", factor_volatilities)
        object.__setattr__(self, "mean_shift_bps", _to_decimal(self.mean_shift_bps))
        if not self.scenarios:
            raise ValueError("MultiFactorDeliveryOptionModel requires at least one scenario.")
        normalized_scenarios = tuple(
            scenario if isinstance(scenario, MultiFactorScenario) else MultiFactorScenario(tuple(scenario))  # type: ignore[arg-type]
            for scenario in self.scenarios
        )
        for scenario in normalized_scenarios:
            if len(scenario.factor_moves) != len(factor_volatilities):
                raise ValueError("Each scenario factor_moves tuple must match factor_volatilities_bps length.")
        object.__setattr__(self, "scenarios", _normalized_scenarios(normalized_scenarios))
        normalized_loadings = {
            InstrumentId.parse(key): tuple(_to_decimal(item) for item in value)
            for key, value in self.deliverable_factor_loadings.items()
        }
        for loadings in normalized_loadings.values():
            if len(loadings) != len(factor_volatilities):
                raise ValueError("Each deliverable factor-loading tuple must match factor_volatilities_bps length.")
        object.__setattr__(self, "deliverable_factor_loadings", normalized_loadings)

    def _instrument_shift(
        self,
        instrument_id: InstrumentId,
        *,
        base_shift: Decimal,
        scenario: MultiFactorScenario,
    ) -> Decimal:
        """Return the scenario yield shift for one deliverable."""
        loadings = self.deliverable_factor_loadings.get(instrument_id)
        if loadings is None:
            raise KeyError(f"Missing factor loadings for deliverable {instrument_id}.")
        shift = base_shift + self.mean_shift_bps
        for factor_volatility, factor_move, loading in zip(
            self.factor_volatilities_bps,
            scenario.factor_moves,
            loadings,
            strict=True,
        ):
            shift += factor_volatility * factor_move * loading
        return shift

    def delivery_option_adjustment(
        self,
        contract: GovernmentBondFuture,
        basket: DeliverableBasket,
        *,
        base_yield_shift_bps: object = Decimal(0),
        prefer_published_conversion_factor: bool = True,
        pricer: BondPricer | None = None,
    ) -> DeliveryOptionResult:
        """Return the probability-weighted delivery-option adjustment."""
        resolved_pricer = pricer or BondPricer()
        base_shift = _to_decimal(base_yield_shift_bps)
        baseline_candidates = _scenario_equivalent_prices(
            contract,
            basket,
            base_yield_shift_bps=base_shift,
            prefer_published_conversion_factor=prefer_published_conversion_factor,
            pricer=resolved_pricer,
        )
        baseline = baseline_candidates[0]
        weighted_adjustment = Decimal(0)
        scenarios: list[DeliveryOptionScenario] = []
        for index, scenario in enumerate(self.scenarios):
            shifts = {
                deliverable.instrument_id: self._instrument_shift(
                    deliverable.instrument_id,
                    base_shift=base_shift,
                    scenario=scenario,
                )
                for deliverable in basket.deliverables
            }
            scenario_candidates = _scenario_candidates_with_instrument_shifts(
                contract,
                basket,
                yield_shifts_bps=shifts,
                prefer_published_conversion_factor=prefer_published_conversion_factor,
                pricer=resolved_pricer,
            )
            forced_baseline = next(item for item in scenario_candidates if item.instrument_id == baseline.instrument_id)
            optimal = scenario_candidates[0]
            switching_benefit = max(
                forced_baseline.futures_equivalent_price - optimal.futures_equivalent_price,
                Decimal(0),
            )
            average_shift = sum(shifts.values(), start=Decimal(0)) / Decimal(len(shifts))
            weighted_adjustment += scenario.probability * switching_benefit
            scenarios.append(
                DeliveryOptionScenario(
                    yield_shift_bps=average_shift - base_shift,
                    cheapest_to_deliver=optimal.instrument_id,
                    forced_baseline_equivalent_price=forced_baseline.futures_equivalent_price,
                    optimal_equivalent_price=optimal.futures_equivalent_price,
                    switching_benefit=switching_benefit,
                    probability=scenario.probability,
                    scenario_label=scenario.label or f"multi_factor_{index}",
                )
            )
        return DeliveryOptionResult(
            baseline_ctd=baseline.instrument_id,
            baseline_futures_equivalent_price=baseline.futures_equivalent_price,
            delivery_option_adjustment=weighted_adjustment,
            scenarios=tuple(scenarios),
        )


__all__ = ["MultiFactorDeliveryOptionModel", "MultiFactorScenario"]



"""Zero-coupon bond instruments."""

from __future__ import annotations

from dataclasses import dataclass, field
from decimal import Decimal

from fuggers_py.core.types import Currency, Date, Frequency

from ..cashflows import CashFlowGenerator, Schedule, ScheduleConfig
from ..errors import InvalidBondSpec
from ..traits import Bond, BondAnalytics, BondCashFlow
from ..types import BondIdentifiers, YieldCalculationRules


@dataclass(frozen=True, slots=True)
class ZeroCouponBond(BondAnalytics, Bond):
    """Bond that pays principal only at maturity."""

    _issue_date: Date
    _maturity_date: Date
    _currency: Currency = Currency.USD
    _notional: Decimal = Decimal(100)
    _identifiers: BondIdentifiers = field(default_factory=BondIdentifiers)
    _rules: YieldCalculationRules = field(default_factory=YieldCalculationRules.us_treasury)

    _schedule: Schedule = field(init=False, repr=False)
    _cashflows: list[BondCashFlow] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        if self._maturity_date <= self._issue_date:
            raise InvalidBondSpec(reason="maturity_date must be after issue_date.")
        if self._notional <= 0:
            raise InvalidBondSpec(reason="notional must be positive.")

        rules = self.rules()
        schedule_config = ScheduleConfig(
            start_date=self._issue_date,
            end_date=self._maturity_date,
            frequency=Frequency.ZERO,
            calendar=rules.calendar,
            business_day_convention=rules.business_day_convention,
            end_of_month=rules.end_of_month,
            stub_rules=rules.stub_rules,
        )
        schedule = Schedule.generate(schedule_config)
        cashflows = CashFlowGenerator.fixed_rate_bond_cashflows(
            schedule=schedule,
            coupon_rate=Decimal(0),
            notional=self._notional,
            rules=rules,
        )
        object.__setattr__(self, "_schedule", schedule)
        object.__setattr__(self, "_cashflows", cashflows)

    def identifiers(self) -> BondIdentifiers:
        return self._identifiers

    def currency(self) -> Currency:
        return self._currency

    def notional(self) -> Decimal:
        return self._notional

    def issue_date(self) -> Date:
        return self._issue_date

    def maturity_date(self) -> Date:
        return self._maturity_date

    def frequency(self) -> Frequency:
        return Frequency.ZERO

    def rules(self) -> YieldCalculationRules:
        return self._rules

    def cash_flows(self, from_date: Date | None = None) -> list[BondCashFlow]:
        return CashFlowGenerator.future_cashflows(self._cashflows, from_date)

    def accrued_interest(self, settlement_date: Date) -> Decimal:
        """Zero-coupon bonds do not accrue interest between issuance and maturity."""
        return Decimal(0)


__all__ = ["ZeroCouponBond"]



"""Credit-default swap instruments."""

from __future__ import annotations

from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum

from fuggers_py.bonds.cashflows import Schedule, ScheduleConfig
from fuggers_py.bonds.types import CalendarId, StubPeriodRules
from fuggers_py.core.calendars import BusinessDayConvention
from fuggers_py.core.daycounts import DayCountConvention
from fuggers_py.core.types import Currency, Date, Frequency
from fuggers_py.data.ids import InstrumentId


def _to_decimal(value: object) -> Decimal:
    if isinstance(value, Decimal):
        return value
    return Decimal(str(value))


def _coerce_currency(value: Currency | str) -> Currency:
    if isinstance(value, Currency):
        return value
    return Currency.from_code(str(value))


def _coerce_frequency(value: Frequency | str) -> Frequency:
    if isinstance(value, Frequency):
        return value
    normalized = value.strip().upper().replace("-", "_").replace(" ", "_")
    aliases = {
        "ANNUAL": Frequency.ANNUAL,
        "YEARLY": Frequency.ANNUAL,
        "SEMIANNUAL": Frequency.SEMI_ANNUAL,
        "SEMI_ANNUAL": Frequency.SEMI_ANNUAL,
        "SEMI": Frequency.SEMI_ANNUAL,
        "QUARTERLY": Frequency.QUARTERLY,
        "QUARTER": Frequency.QUARTERLY,
        "MONTHLY": Frequency.MONTHLY,
        "MONTH": Frequency.MONTHLY,
        "ZERO": Frequency.ZERO,
    }
    if normalized in aliases:
        return aliases[normalized]
    return Frequency[normalized]


def _coerce_day_count(value: DayCountConvention | str) -> DayCountConvention:
    if isinstance(value, DayCountConvention):
        return value
    normalized = value.strip().upper().replace("/", "_")
    aliases = {
        "ACT360": DayCountConvention.ACT_360,
        "ACT_360": DayCountConvention.ACT_360,
        "ACT365F": DayCountConvention.ACT_365_FIXED,
        "ACT_365_FIXED": DayCountConvention.ACT_365_FIXED,
        "ACT365L": DayCountConvention.ACT_365_LEAP,
        "ACT_365_LEAP": DayCountConvention.ACT_365_LEAP,
        "30E_360": DayCountConvention.THIRTY_360_E,
        "30_360_E": DayCountConvention.THIRTY_360_E,
        "30_360_US": DayCountConvention.THIRTY_360_US,
    }
    if normalized in aliases:
        return aliases[normalized]
    return DayCountConvention[normalized]


def _coerce_calendar(value: CalendarId | str) -> CalendarId:
    if isinstance(value, CalendarId):
        return value
    return CalendarId.new(value)


def _coerce_business_day_convention(value: BusinessDayConvention | str) -> BusinessDayConvention:
    if isinstance(value, BusinessDayConvention):
        return value
    normalized = value.strip().upper().replace("-", "_").replace(" ", "_")
    return BusinessDayConvention[normalized]


class ProtectionSide(str, Enum):
    """Protection direction.

    Sign convention:
    - `BUY` means the holder buys protection, pays running premium, and profits
      when the credit curve widens or defaults occur.
    - `SELL` is the opposite economic position.
    """

    BUY = "BUY"
    SELL = "SELL"

    @classmethod
    def parse(cls, value: "ProtectionSide" | str) -> "ProtectionSide":
        """Parse a protection-side value from a string or enum instance."""

        if isinstance(value, cls):
            return value
        normalized = value.strip().upper().replace("-", "_").replace(" ", "_")
        aliases = {
            "BUY": cls.BUY,
            "BUY_PROTECTION": cls.BUY,
            "PROTECTION_BUYER": cls.BUY,
            "SELL": cls.SELL,
            "SELL_PROTECTION": cls.SELL,
            "PROTECTION_SELLER": cls.SELL,
        }
        try:
            return aliases[normalized]
        except KeyError as exc:
            raise ValueError(f"Unsupported protection side: {value!r}.") from exc

    def sign(self) -> Decimal:
        """Return ``+1`` for BUY protection and ``-1`` for SELL protection."""

        return Decimal(1) if self is ProtectionSide.BUY else Decimal(-1)

    def opposite(self) -> "ProtectionSide":
        """Return the opposite protection side."""

        return ProtectionSide.SELL if self is ProtectionSide.BUY else ProtectionSide.BUY


@dataclass(frozen=True, slots=True)
class CdsPremiumPeriod:
    """Single CDS premium-accrual period and payment date."""

    start_date: Date
    end_date: Date
    payment_date: Date
    year_fraction: Decimal

    def __post_init__(self) -> None:
        object.__setattr__(self, "year_fraction", _to_decimal(self.year_fraction))


@dataclass(frozen=True, slots=True)
class CreditDefaultSwap:
    """Vanilla single-name CDS contract.

    Notes
    -----
    ``running_spread`` and ``upfront`` are raw decimals. Positive upfront is
    paid by the protection buyer under the library's sign convention.
    """

    effective_date: Date
    maturity_date: Date
    running_spread: Decimal
    notional: Decimal = Decimal(1_000_000)
    protection_side: ProtectionSide | str = ProtectionSide.BUY
    recovery_rate: Decimal = Decimal("0.4")
    currency: Currency | str = Currency.USD
    payment_frequency: Frequency | str = Frequency.QUARTERLY
    day_count_convention: DayCountConvention | str = DayCountConvention.ACT_360
    calendar: CalendarId | str = field(default_factory=CalendarId.weekend_only)
    business_day_convention: BusinessDayConvention | str = BusinessDayConvention.MODIFIED_FOLLOWING
    end_of_month: bool = True
    stub_rules: StubPeriodRules = field(default_factory=StubPeriodRules.default)
    accrued_on_default_fraction: Decimal = Decimal("0.5")
    upfront: Decimal = Decimal(0)
    settlement_date: Date | None = None
    reference_entity: str | None = None
    instrument_id: InstrumentId | None = None

    def __post_init__(self) -> None:
        object.__setattr__(self, "running_spread", _to_decimal(self.running_spread))
        object.__setattr__(self, "notional", _to_decimal(self.notional))
        object.__setattr__(self, "protection_side", ProtectionSide.parse(self.protection_side))
        object.__setattr__(self, "recovery_rate", _to_decimal(self.recovery_rate))
        object.__setattr__(self, "currency", _coerce_currency(self.currency))
        object.__setattr__(self, "payment_frequency", _coerce_frequency(self.payment_frequency))
        object.__setattr__(self, "day_count_convention", _coerce_day_count(self.day_count_convention))
        object.__setattr__(self, "calendar", _coerce_calendar(self.calendar))
        object.__setattr__(
            self,
            "business_day_convention",
            _coerce_business_day_convention(self.business_day_convention),
        )
        object.__setattr__(self, "accrued_on_default_fraction", _to_decimal(self.accrued_on_default_fraction))
        object.__setattr__(self, "upfront", _to_decimal(self.upfront))
        if self.reference_entity is not None:
            object.__setattr__(self, "reference_entity", self.reference_entity.strip())
        if self.instrument_id is not None:
            object.__setattr__(self, "instrument_id", InstrumentId.parse(self.instrument_id))

        if self.maturity_date <= self.effective_date:
            raise ValueError("CreditDefaultSwap requires maturity_date after effective_date.")
        if self.notional <= Decimal(0):
            raise ValueError("CreditDefaultSwap notional must be positive.")
        if self.payment_frequency.is_zero():
            raise ValueError("CreditDefaultSwap payment_frequency must be non-zero.")
        if self.recovery_rate < Decimal(0) or self.recovery_rate >= Decimal(1):
            raise ValueError("CreditDefaultSwap recovery_rate must lie in [0, 1).")
        if self.accrued_on_default_fraction < Decimal(0) or self.accrued_on_default_fraction > Decimal(1):
            raise ValueError("accrued_on_default_fraction must lie in [0, 1].")

    def schedule(self) -> Schedule:
        """Return the premium payment schedule for the CDS."""

        return Schedule.generate(
            ScheduleConfig(
                start_date=self.effective_date,
                end_date=self.maturity_date,
                frequency=self.payment_frequency,
                calendar=self.calendar,
                business_day_convention=self.business_day_convention,
                end_of_month=self.end_of_month,
                stub_rules=self.stub_rules,
            )
        )

    def premium_periods(self) -> tuple[CdsPremiumPeriod, ...]:
        """Return premium periods with accrual factors and payment dates."""

        schedule = self.schedule()
        day_count = self.day_count_convention.to_day_count()
        periods: list[CdsPremiumPeriod] = []
        for index in range(1, len(schedule.unadjusted_dates)):
            accrual_start = schedule.unadjusted_dates[index - 1]
            accrual_end = schedule.unadjusted_dates[index]
            periods.append(
                CdsPremiumPeriod(
                    start_date=accrual_start,
                    end_date=accrual_end,
                    payment_date=schedule.dates[index],
                    year_fraction=day_count.year_fraction(accrual_start, accrual_end),
                )
            )
        return tuple(periods)

    def loss_given_default(self) -> Decimal:
        """Return ``1 - recovery_rate``."""

        return Decimal(1) - self.recovery_rate

    def upfront_amount(self) -> Decimal:
        """Return upfront payment in currency units."""

        return self.notional * self.upfront


Cds = CreditDefaultSwap


__all__ = [
    "Cds",
    "CdsPremiumPeriod",
    "CreditDefaultSwap",
    "ProtectionSide",
]
